! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module slib_bary
  
  use slib_constants
  
  implicit none
  
  private
  public calc_bary

contains
  
  subroutine calc_bary(ARC_X, ARC_Y, myModelPSF, COEFF1, COEFF2, IE)
    
    implicit none

    double precision, INTENT(IN) :: ARC_X, ARC_Y    
    TYPE(MODELPSF), INTENT(IN) :: myModelPSF
    integer, intent(in) :: IE  
    double precision, intent(out) :: COEFF1(:), COEFF2(:)

    integer, DIMENSION (3) :: ORDER
    integer, DIMENSION (3) :: WEIGHT
    integer I, IAPEX

    write(*,*) "QUERING:",ARC_X, ARC_Y, IE

    call findNnearest(ARC_X, ARC_Y, myModelPSF, ORDER, WEIGHT)
    call get_weights(ARC_X, ARC_Y, myModelPSF, ORDER, WEIGHT)


    write(*,*) WEIGHT


    
    COEFF1 = 0.D0
    COEFF2 = 0.D0
    

    do IAPEX = 1, SIZE(ORDER)

       I = ORDER(IAPEX)
       
       COEFF1 =  COEFF1 + WEIGHT(IAPEX) * myModelPSF%C1(:,I,IE)
       COEFF2 =  COEFF2 + WEIGHT(IAPEX) * myModelPSF%C2(:,I,IE)
       

       write(*,*) WEIGHT(IAPEX) 
       write(*,*) myModelPSF%C1(:,I,IE)
    enddo
    COEFF1 = COEFF1 / SUM(WEIGHT)
    COEFF2 = COEFF2 / SUM(WEIGHT)

  end subroutine calc_bary



  subroutine findNnearest(ARC_X, ARC_Y, myModelPSF, ORDER, WEIGHT)
    
    use quicksort
      
    implicit none
    
    double precision, INTENT(IN) :: ARC_X, ARC_Y    
    TYPE(MODELPSF), INTENT(IN) :: myModelPSF
         
    double precision, DIMENSION (SIZE(myModelPSF%XOFF)) :: DR
    integer, DIMENSION (:) :: ORDER
    integer, DIMENSION (:) :: WEIGHT

    !local variables
    integer I
    double precision dx, dy

    do I=1,SIZE(myModelPSF%XOFF)
       
       dx = (myModelPSF%XOFF(i) -  ARC_X)
       dy = (myModelPSF%YOFF(i) -  ARC_Y)
       DR(i) = dx * dx + dy * dy
    enddo

    call quick_sort3(DR, ORDER)
    
    !write(*,*) "findNnearest:",DR(1:3)
    !write(*,*) "findNnearest:",ORDER
    !do I = 1,SIZE(ORDER)
    !  write(*,*) "findNnearest",I, ORDER(I), DR(I), &
    !       myModelPSF%XOFF(ORDER(I)), myModelPSF%YOFF(ORDER(I))
    !enddo
  end subroutine findNnearest
  

  subroutine quick_sort3(dr, order)

    implicit none

    double precision, DIMENSION (:) :: DR
    integer, DIMENSION (:) :: ORDER

    integer N, I, ISWAP(1),ISWAP1
    double precision TEMP


    N = SIZE(DR)

    do I=1, SIZE(ORDER)
       
       ISWAP=MINLOC(DR(I:N))

       ISWAP1=ISWAP(1)+I-1

       !write(*,*) I, N, DR(ISWAP1), ISWAP1, MINVAL(DR(I:N), DIM=1)

       !ISWAP1=ISWAP(1)

       ORDER(I) = ISWAP1
       
       TEMP=DR(I)
       DR(I)=DR(ISWAP1)
       DR(ISWAP1)=TEMP

    enddo

    
  end subroutine quick_sort3

  subroutine get_weights(ARC_X, ARC_Y, myModelPSF, ORDER, WEIGHT)
    
    implicit none

    integer IAPEX, JAPEX, KAPEX, I, J

    double precision, INTENT(IN) :: ARC_X, ARC_Y    
    TYPE(MODELPSF), INTENT(IN) :: myModelPSF        
    integer, DIMENSION (:) :: ORDER
    integer, DIMENSION (SIZE(ORDER)) :: WEIGHT

    double precision area, dr_apexes, &
         dr_point_IAPEX, dr_point_JAPEX

    KAPEX = 4

    do IAPEX = 1, 2
       
       I = ORDER(IAPEX)
       
       do JAPEX=IAPEX+1,3
          
          KAPEX = KAPEX - 1

          J = ORDER(JAPEX)

          dr_apexes = &
               calc_dr( myModelPSF%XOFF(I), myModelPSF%YOFF(I), &
               myModelPSF%XOFF(J), myModelPSF%YOFF(J) )
          dr_point_IAPEX = &
               calc_dr( myModelPSF%XOFF(I), myModelPSF%YOFF(I), &
                ARC_X, ARC_Y)

          dr_point_JAPEX = &
               calc_dr( myModelPSF%XOFF(J), myModelPSF%YOFF(J), &
                ARC_X, ARC_Y)

          area = calc_heron(dr_apexes, dr_point_IAPEX, dr_point_JAPEX)
          !write(*,*) 
          !write(*,*) "GETWEIGHTS STARTING"
          !write(*,*) "GETWEIGHTS I:", IAPEX, myModelPSF%XOFF(I), myModelPSF%YOFF(I)
          !write(*,*) "GETWEIGHTS J:", JAPEX, myModelPSF%XOFF(J), myModelPSF%YOFF(J)
          !write(*,*) "GETWEIGHTS DR I-J:", dr_apexes
          !write(*,*) "GETWEIGHTS DR I-POINT:",dr_point_IAPEX
          !write(*,*) "GETWEIGHTS DR J-POINT:",dr_point_JAPEX
          !write(*,*) "GETWEIGHTS AREA:",KAPEX,area
          
          weight(KAPEX) = area

       enddo
    enddo
          

  end subroutine get_weights


  double precision function calc_heron (dr_apexes, dr_point_IAPEX, dr_point_JAPEX) 

    implicit none

    double precision, INTENT(IN) :: dr_apexes, dr_point_IAPEX, dr_point_JAPEX

    double precision p

    p = 0.5 * ( dr_apexes + dr_point_IAPEX + dr_point_JAPEX )
    
    calc_heron = sqrt(p * (p - dr_apexes ) * (p - dr_point_IAPEX ) * (p - dr_point_JAPEX))


  end function calc_heron


  double precision function calc_dr (x1, y1, x2, y2) 

    implicit none

    double precision, INTENT(IN) :: x1, x2, y1, y2

    
    calc_dr= sqrt((x1-x2)*(x1-x2) + (y1-y2) * (y1-y2))


  end function calc_dr

end module slib_bary
