! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module slib_calc_space

  use slib_constants

  implicit none
        
  private 
  ! ** GL:  (public :: get_single_event)
  ! ** AGE 14/03/17:  public :: lib_get_xy
  public :: slib_assign, slib_assign_heal, &
       polar_normalise_unity, polar_rotate, &
       slib_fluxR, get_single_event, slib_get_xy, &
       slib_fluxEEF

contains


  subroutine slib_assign(myEVENTFILE, mySRC, myModelPSF, myCPSF, myStackPSF)

    use slib_basic_calcs
    !use  slib_bary

    implicit none

    TYPE(EVENTFILE), intent(in) :: myEVENTFILE

    TYPE(SRC), intent(in) :: mySRC
        
    TYPE(MODELPSF), intent(in) :: myModelPSF(:)
    TYPE(CPSF), intent(in) :: myCPSF


    TYPE(STACKPSF) myStackPSF

    !OLD
    !double precision COEFF1_TEMP (SIZE( myModelPSF(1)%NX1)),  COEFF2_TEMP (SIZE( myModelPSF(1)%NX2)) 
    !double precision COEFF1_TEMP1 (SIZE( myModelPSF(1)%NX1)),  COEFF2_TEMP1 (SIZE( myModelPSF(1)%NX2)) 
    !complex PCOEFF1_TEMP (SIZE( myModelPSF(1)%NX1)),  PCOEFF2_TEMP (SIZE( myModelPSF(1)%NX2)) 
    !complex CCOEFF1_TEMP (SIZE( myModelPSF(1)%NX1)),  CCOEFF2_TEMP (SIZE( myModelPSF(1)%NX2)) 
    !NEW
    double precision  COEFF0_TEMP(SIZE(myCPSF%NX0)), &
         COEFF1_TEMP(SIZE(myCPSF%NX1)),  COEFF2_TEMP(SIZE(myCPSF%NX2)) 
    double precision COEFF0_TEMP1(SIZE(myCPSF%NX0)), &
         COEFF1_TEMP1(SIZE(myCPSF%NX1)),  COEFF2_TEMP1(SIZE(myCPSF%NX2)) 
    complex PCOEFF0_TEMP (SIZE( myCPSF%NX0)), PCOEFF1_TEMP (SIZE( myCPSF%NX1)), &
         PCOEFF2_TEMP (SIZE( myCPSF%NX2)) 
    complex CCOEFF0_TEMP (SIZE( myCPSF%NX0)),  CCOEFF1_TEMP (SIZE( myCPSF%NX1)), &
         CCOEFF2_TEMP (SIZE( myCPSF%NX2)) 

    integer NP, ievt, isrc

    double precision theta, x1, y1, z1, x2, y2, z2

    double precision norm1, norm2, energy

    !these are for data/time, benchmarking
    character(8)  :: date
    character(10) :: time
    character(5)  :: zone
    integer,dimension(8) :: values
    INTEGER :: &
         nb_ticks_initial, & ! initial value of the clock tick counter
         nb_ticks_final,   & ! final value of the clock tick counter
         nb_ticks_max,     & ! maximum value of the clock counter
         nb_ticks_sec,     & ! number of clock ticks per second
         nb_ticks           ! number of clock ticks of the code
    REAL :: elapsed_time  ! real time in seconds
    !these are for data/time, benchmarking



    !initialise arrays of stacked PSF
    myStackPSF%C0=0.0
    myStackPSF%C1=0.0
    myStackPSF%C2=0.0

    !CALL SYSTEM_CLOCK(COUNT_RATE=nb_ticks_sec, COUNT_MAX=nb_ticks_max)
    
    do isrc=1, mySRC%TOTAL
    !do isrc=48, 48
       
       NP=0

       !write(*,*) "Starting with SRC", isrc, mySRC%RA(isrc), mySRC%DEC(isrc), mySRC%RADIUS(isrc) ,  mySRC%TOTAL
       !call date_and_time(date,time,zone,values)
       !write(*,*)"starting on Date: ", date, " Time: ",time
       !write(*,*)
       !write(92,*) "fk5;circle(",mySRC%RA(isrc), ",", mySRC%DEC(isrc),",32i)#text={",isrc,"}"
       !CALL SYSTEM_CLOCK(COUNT=nb_ticks_initial)

       COEFF0_TEMP=0
       COEFF1_TEMP=0
       COEFF2_TEMP=0
       
       x2=mySRC%XD(isrc)
       y2=mySRC%YD(isrc)
       z2=mySRC%ZD(isrc)

       do ievt=1,myEVENTFILE%EVENT%SIZE
          
          x1=myEVENTFILE%EVENT%XD(ievt)
          y1=myEVENTFILE%EVENT%YD(ievt)
          z1=myEVENTFILE%EVENT%ZD(ievt)
          
          
          call slib_xyz2theta(x1, y1, z1, x2, y2, z2, theta)
          !if(ISRC.eq.21)write(*,*) NP,myEVENTFILE%EVENT%RA(ievt), myEVENTFILE%EVENT%DEC(ievt),&
          !     theta, mySRC%RADIUS(isrc)
          if(theta.le.mySRC%RADIUS(isrc))then
             
             energy= myEVENTFILE%EVENT%ENERGY(IEVT)
             
             ! this if statement is only because SIXTE produces
             ! negative energies. 
             if(energy.gt.0 .AND. energy.lt.15000.)then
                !write(92,*) "fk5;box(",myEVENTFILE%EVENT%RA(ievt), ",", myEVENTFILE%EVENT%DEC(ievt),",1i,1i,0)#text={",isrc,"}"

                NP=NP+1
               if(ISRC.eq.48)write(96,*) NP, myEVENTFILE%EVENT%RA(ievt), myEVENTFILE%EVENT%DEC(ievt),&
                    myEVENTFILE%EVENT%ENERGY(IEVT), myEVENTFILE%EVENT%ROLL(IEVT),myEVENTFILE%EVENT%RAWX(IEVT),&
                    myEVENTFILE%EVENT%RAWY(IEVT)
                         
                         
                call get_single_event(myEVENTFILE, myModelPSF, myCPSF, &
                     IEVT, COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP)                
                
                myStackPSF%C0(:, ISRC) = myStackPSF%C0(:, ISRC) + COEFF0_TEMP                
                myStackPSF%C1(:, ISRC) = myStackPSF%C1(:, ISRC) + COEFF1_TEMP
                myStackPSF%C2(:, ISRC) = myStackPSF%C2(:, ISRC) + COEFF2_TEMP

                !write(*,*) energy, NP, mySRC%RADIUS(isrc)
                
             endif
             
          endif
          
       enddo
       
       ! I normlise PSF by dividing by the number of photons
       ! within the searched radius. 
       myStackPSF%C0(:, ISRC) = myStackPSF%C0(:, ISRC) /  float(NP)
       myStackPSF%C1(:, ISRC) = myStackPSF%C1(:, ISRC) /  float(NP)
       myStackPSF%C2(:, ISRC) = myStackPSF%C2(:, ISRC) /  float(NP)

       !write(*,*) NP, ISRC, mySRC%RADIUS(isrc)
      ! testing normalisation
      ! CCOEFF1_TEMP = CMPLX( myStackPSF%C1(:, ISRC), 0)
      ! CCOEFF2_TEMP = CMPLX( myStackPSF%C2(:, ISRC), 0)
      ! 
      ! ! go to polar coefficients
      ! PCOEFF1_TEMP=matmul(TRANSPOSE(myModelPSF(itel)%MATRIX_INV1), CCOEFF1_TEMP)
      ! PCOEFF2_TEMP=matmul(TRANSPOSE(myModelPSF(itel)%MATRIX_INV2), CCOEFF2_TEMP)  
      !
      ! call polar_normalise_unity(SIZE( myModelPSF(itel)%NX1), PCOEFF1_TEMP, myModelPSF(itel)%PN1, & 
      !      myModelPSF(itel)%PM1, myModelPSF(itel)%BETA1, NORM1)
      ! call polar_normalise_unity(SIZE( myModelPSF(itel)%NX2), PCOEFF2_TEMP, myModelPSF(itel)%PN2, &
      !      myModelPSF(itel)%PM2 , myModelPSF(itel)%BETA2, NORM2)             
             

       !write(*,*) "Src:", isrc, " photons", NP, "NORMS:", NORM1, NORM2, NORM1 + NORM2
       !CALL SYSTEM_CLOCK(COUNT=nb_ticks_final)
       !nb_ticks = nb_ticks_final - nb_ticks_initial
       !IF (nb_ticks_final < nb_ticks_initial) &
       !     nb_ticks = nb_ticks + nb_ticks_max
       !elapsed_time   = REAL(nb_ticks) / nb_ticks_sec     
       !write(*,*)  "Src:", isrc, " photons/elapsed time", NP, elapsed_time 

    enddo
    !stop
  end subroutine slib_assign
  



  subroutine slib_assign_heal(myEVENTFILE, mySRC, myModelPSF, myCPSF, myStackPSF)

    use slib_basic_calcs
    use slib_heal
    use slib_data_evt, ONLY: slib_sort_heal
    !use  slib_bary

    implicit none

    TYPE(EVENTFILE), intent(in) :: myEVENTFILE

    TYPE(SRC), intent(in) :: mySRC
        
    TYPE(MODELPSF), intent(in) :: myModelPSF(:)
    TYPE(CPSF), intent(in) :: myCPSF


    TYPE(STACKPSF) myStackPSF

    !OLD
    !double precision COEFF1_TEMP (SIZE( myModelPSF(1)%NX1)),  COEFF2_TEMP (SIZE( myModelPSF(1)%NX2)) 
    !double precision COEFF1_TEMP1 (SIZE( myModelPSF(1)%NX1)),  COEFF2_TEMP1 (SIZE( myModelPSF(1)%NX2)) 
    !complex PCOEFF1_TEMP (SIZE( myModelPSF(1)%NX1)),  PCOEFF2_TEMP (SIZE( myModelPSF(1)%NX2)) 
    !complex CCOEFF1_TEMP (SIZE( myModelPSF(1)%NX1)),  CCOEFF2_TEMP (SIZE( myModelPSF(1)%NX2)) 
    !NEW
    double precision  COEFF0_TEMP(SIZE(myCPSF%NX0)), &
         COEFF1_TEMP(SIZE(myCPSF%NX1)),  COEFF2_TEMP(SIZE(myCPSF%NX2)) 
    double precision COEFF0_TEMP1(SIZE(myCPSF%NX0)), &
         COEFF1_TEMP1(SIZE(myCPSF%NX1)),  COEFF2_TEMP1(SIZE(myCPSF%NX2)) 
    complex PCOEFF0_TEMP (SIZE( myCPSF%NX0)), PCOEFF1_TEMP (SIZE( myCPSF%NX1)), &
         PCOEFF2_TEMP (SIZE( myCPSF%NX2)) 
    complex CCOEFF0_TEMP (SIZE( myCPSF%NX0)),  CCOEFF1_TEMP (SIZE( myCPSF%NX1)), &
         CCOEFF2_TEMP (SIZE( myCPSF%NX2)) 

    integer NP, ievt, isrc, i1, i2

    double precision theta, x1, y1, z1, x2, y2, z2

    double precision norm1, norm2, energy

    !these are for data/time, benchmarking
    character(8)  :: date
    character(10) :: time
    character(5)  :: zone
    integer,dimension(8) :: values
    INTEGER :: &
         nb_ticks_initial, & ! initial value of the clock tick counter
         nb_ticks_final,   & ! final value of the clock tick counter
         nb_ticks_max,     & ! maximum value of the clock counter
         nb_ticks_sec,     & ! number of clock ticks per second
         nb_ticks           ! number of clock ticks of the code
    REAL :: elapsed_time  ! real time in seconds
    !these are for data/time, benchmarking

    ! relate to heal
    integer ipix, nlist, ipix_running
    double precision search_radius_rad
    integer, dimension(10000) :: listpix

    
    ! check is event are sorted by healpix pixel (position)
    ! if not then apply sorting.
   ! Call sorting routine only, if we have more than zero events  (GL 10/02/2020) :
    if(.NOT.myEVENTFILE%EVENT%HEALSORT .AND. myEVENTFILE%EVENT%SIZE > 0) call slib_sort_heal(myEVENTFILE)
    
    !initialise arrays of stacked PSF
    myStackPSF%C0=0.0
    myStackPSF%C1=0.0
    myStackPSF%C2=0.0

    !CALL SYSTEM_CLOCK(COUNT_RATE=nb_ticks_sec, COUNT_MAX=nb_ticks_max)

   
    
    do isrc=1, mySRC%TOTAL

       NP=0

       x2=mySRC%XD(isrc)
       y2=mySRC%YD(isrc)
       z2=mySRC%ZD(isrc)
                   
       !write(*,*) "Starting with SRC", isrc, mySRC%RA(isrc), mySRC%DEC(isrc), mySRC%RADIUS(isrc) ,  mySRC%TOTAL
       !call date_and_time(date,time,zone,values)
       !write(*,*)"starting on Date: ", date, " Time: ",time
       !write(*,*)
       !CALL SYSTEM_CLOCK(COUNT=nb_ticks_initial)

       COEFF0_TEMP=0
       COEFF1_TEMP=0
       COEFF2_TEMP=0

       search_radius_rad =mySRC%RADIUS(isrc) * myDEG2RAD / 3600.0
       call SkyToPix(mySRC%RA(isrc), mySRC%DEC(isrc), myEVENTFILE%EVENT%nside, ipix)
       !if(isrc.eq.21)write(*,*) "SRC IS IN HEALPIX", ipix
       call GetPixRad(ipix, search_radius_rad, myEVENTFILE%EVENT%nside, listpix, nlist)
       !write(*,*) ipix, mySRC%RADIUS(isrc)
       !write(*,*) isrc, nlist,listpix(1:nlist)
       do ipix=1,nlist
          
          ipix_running=listpix(ipix)

          i1 = myEVENTFILE%EVENT%index1( ipix_running )
          i2 = myEVENTFILE%EVENT%index2( ipix_running )
             
          !write(*,*) ipix_running, i1, i2 
             
             
          do ievt = i1, i2

             if(ievt.gt.0)then
                
                x1=myEVENTFILE%EVENT%XD(ievt)
                y1=myEVENTFILE%EVENT%YD(ievt)
                z1=myEVENTFILE%EVENT%ZD(ievt)
          
                   
                call slib_xyz2theta(x1, y1, z1, x2, y2, z2, theta)
                
                if(theta.le.mySRC%RADIUS(isrc))then
             
                   energy= myEVENTFILE%EVENT%ENERGY(IEVT)
             
                   ! this if statement is only because SIXTE produces
                   ! negative energies. 
                   if(energy.gt.0 .AND. energy.lt.15000.)then
                         
                      NP=NP+1

                         
                      !if(ISRC.eq.21)write(*,*) NP, myEVENTFILE%EVENT%RA(ievt), myEVENTFILE%EVENT%DEC(ievt),&
                      !     myEVENTFILE%EVENT%ENERGY(IEVT), myEVENTFILE%EVENT%ROLL(IEVT),&
                      !     theta,   mySRC%RADIUS(isrc)
                         
                      call get_single_event(myEVENTFILE, myModelPSF, myCPSF, &
                           IEVT, COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP)
                
                      myStackPSF%C0(:, ISRC) = myStackPSF%C0(:, ISRC) + COEFF0_TEMP                
                      myStackPSF%C1(:, ISRC) = myStackPSF%C1(:, ISRC) + COEFF1_TEMP
                      myStackPSF%C2(:, ISRC) = myStackPSF%C2(:, ISRC) + COEFF2_TEMP
                         
                      !write(*,*) energy, NP, mySRC%RADIUS(isrc)
                
                   endif
             
                endif
                   
             endif

          enddo
          
       enddo
       ! I normlise PSF by dividing by the number of photons
       ! within the searched radius. 
       myStackPSF%C0(:, ISRC) = myStackPSF%C0(:, ISRC) /  float(NP)
       myStackPSF%C1(:, ISRC) = myStackPSF%C1(:, ISRC) /  float(NP)
       myStackPSF%C2(:, ISRC) = myStackPSF%C2(:, ISRC) /  float(NP)

       !write(*,*) NP, ISRC, mySRC%RADIUS(isrc)
       ! testing normalisation
       ! CCOEFF1_TEMP = CMPLX( myStackPSF%C1(:, ISRC), 0)
       ! CCOEFF2_TEMP = CMPLX( myStackPSF%C2(:, ISRC), 0)
       ! 
       ! ! go to polar coefficients
       ! PCOEFF1_TEMP=matmul(TRANSPOSE(myModelPSF(itel)%MATRIX_INV1), CCOEFF1_TEMP)
       ! PCOEFF2_TEMP=matmul(TRANSPOSE(myModelPSF(itel)%MATRIX_INV2), CCOEFF2_TEMP)  
       !
       ! call polar_normalise_unity(SIZE( myModelPSF(itel)%NX1), PCOEFF1_TEMP, myModelPSF(itel)%PN1, & 
       !      myModelPSF(itel)%PM1, myModelPSF(itel)%BETA1, NORM1)
       ! call polar_normalise_unity(SIZE( myModelPSF(itel)%NX2), PCOEFF2_TEMP, myModelPSF(itel)%PN2, &
       !      myModelPSF(itel)%PM2 , myModelPSF(itel)%BETA2, NORM2)             
       
       
       !write(*,*) "Src:", isrc, " photons", NP, "NORMS:", NORM1, NORM2, NORM1 + NORM2
       !CALL SYSTEM_CLOCK(COUNT=nb_ticks_final)
       !nb_ticks = nb_ticks_final - nb_ticks_initial
       !IF (nb_ticks_final < nb_ticks_initial) &
       !     nb_ticks = nb_ticks + nb_ticks_max
       !elapsed_time   = REAL(nb_ticks) / nb_ticks_sec     
       !write(*,*)  "Src:", isrc, " photons/elapsed time", NP, elapsed_time 

    enddo

  end subroutine slib_assign_heal
  






  subroutine get_single_event(myEVENTFILE, myModelPSF, myCPSF, &
       IEVT, COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP)

    use slib_cimage

    implicit none

    TYPE(EVENTFILE), intent(in) :: myEVENTFILE        
    TYPE(MODELPSF), intent(in) :: myModelPSF(:)
    TYPE(CPSF), intent(in) :: myCPSF
    integer, intent(in) :: IEVT
    DOUBLE PRECISION, intent(out) :: COEFF0_TEMP(:)  
    DOUBLE PRECISION, intent(out) :: COEFF1_TEMP(:)  
    DOUBLE PRECISION, intent(out) :: COEFF2_TEMP(:) 

    integer itel, iccdnr, ii, jj
    double precision dx, dy, x0, y0, arc_x, arc_y
    double precision energy, energy_keV
    double precision angle, norm1, norm2

    integer, dimension (2,2) :: IXY
    integer IE, IX, IY, IC, NP, ISRC

    complex  PCOEFF0_TEMP(SIZE(myCPSF%NX0)), &
         PCOEFF1_TEMP(SIZE(myCPSF%NX1)), PCOEFF2_TEMP(SIZE(myCPSF%NX2)) 
    complex CCOEFF0_TEMP(SIZE(myCPSF%NX0)), &
         CCOEFF1_TEMP(SIZE(myCPSF%NX1)), CCOEFF2_TEMP(SIZE(myCPSF%NX2)) 



    NP = 1 
    ISRC = 1

    !get the telescope number
    iccdnr = myEVENTFILE%EVENT%CCDNR(IEVT)
    do jj = 1, SIZE(myEVENTFILE%WHICH_CCDNR)
       if( myEVENTFILE%WHICH_CCDNR(jj) == iccdnr ) itel=jj
    enddo
    
             
    x0 = myEVENTFILE%CCD%RAW_X0(itel) ! check that iact_tel points to the right index
    y0 = myEVENTFILE%CCD%RAW_Y0(itel)
             
    ! AGE 6/16: Georg pointed out that RAWX and RAWY 
    !           should not be rotated since it is 
    !           assumed that PANTER X == ERO-X-AXIS
    !           PANTER Y == ERO-Y-AXIS
    !angle =  myEVENTFILE%EVENT%ROLL(IEVT)  * myDEG2RAD
    ! AGE 05-06-19: RAWX is flipped wrsp to PANTER-X
    ! this follows changes in CCD orientations in SIXTE
    dx = -(myEVENTFILE%EVENT%RAWX(IEVT) - x0)
    dy = myEVENTFILE%EVENT%RAWY(IEVT) - y0

    arc_x = dx * RAW_PIX_DEG * 60.
    arc_y = dy * RAW_PIX_DEG * 60.
    
    energy= myEVENTFILE%EVENT%ENERGY(IEVT)

    energy_keV= energy / 1.e3
    
    !write(*,*) itel, energy, arc_x, arc_y, myEVENTFILE%EVENT%RAWX(IEVT) , x0,  myEVENTFILE%EVENT%RAWY(IEVT), y0
    
    !ONLY FOR TESTING
    !arc_x = 0.
    !arc_y = 0.
    !energy_keV=1.5
    
       
    ! check that everything between limits
    ! with brycentric interpolation this is not needed any more 
    if(arc_x .lt. myModelPSF(itel)%XMIN) ARC_X  = myModelPSF(itel)%XMIN
    if(arc_x .gt. myModelPSF(itel)%XMAX) ARC_X  = myModelPSF(itel)%XMAX - myModelPSF(itel)%XBIN/10.
    if(arc_y .lt. myModelPSF(itel)%YMIN) ARC_Y  = myModelPSF(itel)%YMIN
    if(arc_y .gt. myModelPSF(itel)%YMAX) ARC_Y  = myModelPSF(itel)%YMAX - myModelPSF(itel)%YBIN/10.
    
    !write(*,*) "IN: ARC_X, ARC_Y: ", arc_x, arc_y, energy_keV
    
    if(energy_keV .lt. myModelPSF(itel)%ENERGY(1)) energy_keV  =  myModelPSF(itel)%ENERGY(1)
    if(energy_keV .gt. myModelPSF(itel)%ENERGY( SIZE(myModelPSF(itel)%ENERGY) )) &
         energy_keV  =  myModelPSF(itel)%ENERGY( SIZE(myModelPSF(itel)%ENERGY) )
    
    !write(*,*) "NEW ENERGY IN: ", energy_keV, myModelPSF(itel)%ENERGY(1), &
    !      myModelPSF(itel)%ENERGY( SIZE(myModelPSF(itel)%ENERGY) )
    
    call findXY(arc_x, arc_y, myModelPSF(itel), IXY)
    call findEN(energy_keV, myModelPSF(itel), IE)
    
    !write(*,*) "found one"
    !write(*,*) myEVENTFILE%EVENT%RAWX(IEVT),&
    !     myEVENTFILE%EVENT%RAWY(IEVT)
    !write(*,*) x0, arc_x, RAW_PIX_DEG*60., & 
    !     myEVENTFILE%EVENT%RAWX(IEVT) - x0
    !write(*,*) arc_x, arc_y!, SIZE(myModelPSF(itel)%XOFF), SIZE(myModelPSF(itel)%YOFF)
    !write(*,*) IXY(1,1), IXY(1,2), IXY(2,1), IXY(2,2)
    !write(*,*) "E:",NP,energy_keV, IE, myModelPSF(itel)%ENERGY(IE)
    !write(*,*) "X:",NP,arc_x, IXY(1,1), IXY(2,2), myModelPSF(itel)%XOFF(IXY(1,1)), myModelPSF(itel)%XOFF(IXY(2,2))
    !write(*,*) "Y:",NP,arc_y, myModelPSF(itel)%YOFF(IXY(1,1)), myModelPSF(itel)%YOFF(IXY(2,2))
    
    !16/07/2016: two posibilities. Trilinear interpolation that works in the case of 
    !regular grid, or barycenrtic interpolation, which works even in the case of
    !non-regular grid. 
    ! call to trilinear
    call trilinear( myModelPSF(itel), IXY, IE, ARC_X, ARC_Y, ENERGY_KEV, &
         COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP)
    ! call to barycentric
    !call calc_bary(ARC_X, ARC_Y,myModelPSF(itel), COEFF1_TEMP, COEFF2_TEMP, IE)
    !call calc_bary(ARC_X, ARC_Y,myModelPSF(itel), COEFF1_TEMP1, COEFF2_TEMP1, IE+1)
    !call inter_linear( myModelPSF(itel), IE, ENERGY_KEV, COEFF1_TEMP, COEFF1_TEMP1)
    !call inter_linear( myModelPSF(itel), IE, ENERGY_KEV, COEFF2_TEMP, COEFF2_TEMP1)       
    !write(*,*) "COEFF1",COEFF1_TEMP(1:3)
    !write(*,*) "COEFF2",COEFF2_TEMP(1:3)
    
    
    !! AGE 06/17: I need all that. Not obsolete anymore 
    CCOEFF0_TEMP = CMPLX( COEFF0_TEMP, 0)
    CCOEFF1_TEMP = CMPLX( COEFF1_TEMP, 0)
    CCOEFF2_TEMP = CMPLX( COEFF2_TEMP, 0)
    
    ! go to polar coefficients
    PCOEFF0_TEMP=matmul(TRANSPOSE(myCPSF%MATRIX_INV0), CCOEFF0_TEMP)
    PCOEFF1_TEMP=matmul(TRANSPOSE(myCPSF%MATRIX_INV1), CCOEFF1_TEMP)
    PCOEFF2_TEMP=matmul(TRANSPOSE(myCPSF%MATRIX_INV2), CCOEFF2_TEMP)
    
    ! this angle is in deg and is transformed later on within POLAR_RATATE to RAD
    ! Should do this calc once at the beggining.
    ! AGE:  I have empirically found that I should rotate by the minus angle given 
    ! in the header. Further testing is required here
    angle = -myEVENTFILE%EVENT%ROLL(IEVT)
    ! *** set angle=0 only for testing ***
    !angle=0.
    ! *** set angle=0 only for testing ***
    
    
    
    call polar_rotate(SIZE( myCPSF%NX0), PCOEFF0_TEMP, myCPSF%PN0, & 
         myModelPSF(itel)%PM0 , angle)             
    call polar_rotate(SIZE( myCPSF%NX1), PCOEFF1_TEMP, myCPSF%PN1, & 
         myModelPSF(itel)%PM1 , angle)             
    call polar_rotate(SIZE( myCPSF%NX2), PCOEFF2_TEMP, myCPSF%PN2, &
         myModelPSF(itel)%PM2 , angle)
    
    ! AGE: the functions below are making sure that roated PSF normalise to unity. They should not
    !      be needed because shapelet coefficients are normalised so that the total flux
    !      in the unrotated PSF is unity. Rotation should preserve that.
    !call polar_normalise_unity(SIZE( myModelPSF(itel)%NX1), PCOEFF1_TEMP, myModelPSF(itel)%PN1, & 
    !     myModelPSF(itel)%PM1, myModelPSF(itel)%BETA1, NORM1)
    !call polar_normalise_unity(SIZE( myModelPSF(itel)%NX2), PCOEFF2_TEMP, myModelPSF(itel)%PN2, &
    !     myModelPSF(itel)%PM2 , myModelPSF(itel)%BETA2, NORM2)             
    !PCOEFF1_TEMP = PCOEFF1_TEMP / (NORM1+NORM2)
    !PCOEFF2_TEMP = PCOEFF2_TEMP / (NORM1+NORM2)             
    !
    CCOEFF0_TEMP=matmul(TRANSPOSE(myCPSF%MATRIX0), PCOEFF0_TEMP)
    CCOEFF1_TEMP=matmul(TRANSPOSE(myCPSF%MATRIX1), PCOEFF1_TEMP)
    CCOEFF2_TEMP=matmul(TRANSPOSE(myCPSF%MATRIX2), PCOEFF2_TEMP)
    COEFF0_TEMP = REALPART( CCOEFF0_TEMP)
    COEFF1_TEMP = REALPART( CCOEFF1_TEMP)
    COEFF2_TEMP = REALPART( CCOEFF2_TEMP)
    
    
    !CCOEFF1_TEMP = CMPLX( COEFF1_TEMP, 0)
    !CCOEFF2_TEMP = CMPLX( COEFF2_TEMP, 0)
    !
    !      call slib_write2file(SIZE(myCPSF%NX1), CCOEFF1_TEMP, &
    !           myCPSF%NX1, myCPSF%NY1, &
    !           real(myCPSF%BETA1), &
    !           SIZE(myCPSF%NX2), CCOEFF2_TEMP, myCPSF%NX2, &
    !           myCPSF%NY2, &
    !           real(myCPSF%BETA2), &
    !           100, ISRC, IE, NP)
    !      if(NP>=10)stop
    
    
       
  end subroutine get_single_event



  subroutine slib_get_xy(DX, DY, ENERGY, ROLL, myModelPSF, &
       COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP, NP)

    use slib_cimage
    implicit none

    TYPE(MODELPSF), intent(in) :: myModelPSF
    double precision, intent(in) :: DX, DY, ENERGY, ROLL
    DOUBLE PRECISION, intent(out) :: COEFF0_TEMP(:)  
    DOUBLE PRECISION, intent(out) :: COEFF1_TEMP(:)  
    DOUBLE PRECISION, intent(out) :: COEFF2_TEMP(:) 
    integer, intent(in) :: NP

    double precision arc_x, arc_y
    double precision energy_keV
    double precision angle, norm0, norm1, norm2

    integer, dimension (2,2) :: IXY
    integer IE, IX, IY, IC, ISRC

    complex PCOEFF0_TEMP(SIZE(myModelPSF%NX0)), &
         PCOEFF1_TEMP(SIZE(myModelPSF%NX1)), PCOEFF2_TEMP(SIZE(myModelPSF%NX2)) 
    complex CCOEFF0_TEMP(SIZE(myModelPSF%NX0)), &
         CCOEFF1_TEMP(SIZE( myModelPSF%NX1)),CCOEFF2_TEMP(SIZE( myModelPSF%NX2)) 


    arc_x = DX * RAW_PIX_DEG * 60.
    arc_y = DY * RAW_PIX_DEG * 60.    
    energy_keV= ENERGY / 1.e3

    !ONLY FOR TESTING
    !arc_x = 0.
    !arc_y = 0.
    !energy_keV=4.5
    
       
    ! check that everything between limits
    ! with brycentric interpolation this is not needed any more 
    if(arc_x .lt. myModelPSF%XMIN) ARC_X  = myModelPSF%XMIN
    if(arc_x .gt. myModelPSF%XMAX) ARC_X  = myModelPSF%XMAX - myModelPSF%XBIN/10.
    if(arc_y .lt. myModelPSF%YMIN) ARC_Y  = myModelPSF%YMIN
    if(arc_y .gt. myModelPSF%YMAX) ARC_Y  = myModelPSF%YMAX - myModelPSF%YBIN/10.
    
    !write(*,*) "IN2: ARC_X, ARC_Y: ", arc_x, arc_y
    
    if(energy_keV .lt. myModelPSF%ENERGY(1)) energy_keV  =  myModelPSF%ENERGY(1)
    if(energy_keV .gt. myModelPSF%ENERGY( SIZE(myModelPSF%ENERGY) )) &
         energy_keV  =  myModelPSF%ENERGY( SIZE(myModelPSF%ENERGY) )
    
    !write(*,*) "NEW ENERGY IN: ", energy_keV, myModelPSF%ENERGY(1), &
    !     myModelPSF%ENERGY( SIZE(myModelPSF%ENERGY) )
    
    call findXY(arc_x, arc_y, myModelPSF, IXY)
    call findEN(energy_keV, myModelPSF, IE)
    
    !write(*,*) "found one"
    !write(*,*) myEVENTFILE%EVENT%RAWX(IEVT),&
    !     myEVENTFILE%EVENT%RAWY(IEVT)
    !write(*,*) x0, arc_x, RAW_PIX_DEG*60., & 
    !     myEVENTFILE%EVENT%RAWX(IEVT) - x0
    !write(*,*) arc_x, arc_y!, SIZE(myModelPSF(itel)%XOFF), SIZE(myModelPSF(itel)%YOFF)
    !!write(*,*) IXY(1,1), IXY(1,2), IXY(2,1), IXY(2,2)
    !write(*,*) "E:",NP,energy_keV, IE
    !write(*,*) "X:",NP,arc_x, IXY(1,1), IXY(2,2)
    !write(*,*) "Y:",NP,arc_y, IXY(2,1), IXY(2,2)
    
    !16/07/2016: two posibilities. Trilinear interpolation that works in the case of 
    !regular grid, or barycenrtic interpolation, which works even in the case of
    !non-regular grid. 
    ! call to trilinear
    call trilinear( myModelPSF, IXY, IE, ARC_X, ARC_Y, ENERGY_KEV, &
         COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP)
    ! call to barycentric
    !call calc_bary(ARC_X, ARC_Y,myModelPSF(itel), COEFF1_TEMP, COEFF2_TEMP, IE)
    !call calc_bary(ARC_X, ARC_Y,myModelPSF(itel), COEFF1_TEMP1, COEFF2_TEMP1, IE+1)
    !call inter_linear( myModelPSF(itel), IE, ENERGY_KEV, COEFF1_TEMP, COEFF1_TEMP1)
    !call inter_linear( myModelPSF(itel), IE, ENERGY_KEV, COEFF2_TEMP, COEFF2_TEMP1)       
    !write(*,*) "COEFF1",COEFF1_TEMP(1:3)
    !write(*,*) "COEFF2",COEFF2_TEMP(1:3)
    
    
    !! AGE 06/17: I need all that. Not obsolete anymore 
    CCOEFF0_TEMP = CMPLX( COEFF0_TEMP, 0)
    CCOEFF1_TEMP = CMPLX( COEFF1_TEMP, 0)
    CCOEFF2_TEMP = CMPLX( COEFF2_TEMP, 0)
    
    ! go to polar coefficients
    PCOEFF0_TEMP=matmul(TRANSPOSE(myModelPSF%MATRIX_INV0), CCOEFF0_TEMP)
    PCOEFF1_TEMP=matmul(TRANSPOSE(myModelPSF%MATRIX_INV1), CCOEFF1_TEMP)
    PCOEFF2_TEMP=matmul(TRANSPOSE(myModelPSF%MATRIX_INV2), CCOEFF2_TEMP)
    
    ! this angle is in deg and is transformed later on within POLAR_RATATE to RAD
    ! Should do this calc once at the beggining.
    ! AGE:  I have empirically found that I should rotate by the minus angle given 
    ! in the header. Further testing is required here
    angle = -ROLL
    ! *** set angle=0 only for testing ***
    ! angle=0.
    ! *** set angle=0 only for testing ***
    
    
    call polar_rotate(SIZE(myModelPSF%NX0), PCOEFF0_TEMP, myModelPSF%PN0, & 
         myModelPSF%PM0 , angle)             
    call polar_rotate(SIZE(myModelPSF%NX1), PCOEFF1_TEMP, myModelPSF%PN1, & 
         myModelPSF%PM1 , angle)             
    call polar_rotate(SIZE(myModelPSF%NX2), PCOEFF2_TEMP, myModelPSF%PN2, &
         myModelPSF%PM2 , angle)
    
    ! AGE: the functions below are making sure that roated PSF normalise to unity. They should not
    !      be needed because shapelet coefficients are normalised so that the total flux
    !      in the unrotated PSF is unity. Rotation should preserve that.
    !call polar_normalise_unity(SIZE( myModelPSF(itel)%NX1), PCOEFF1_TEMP, myModelPSF(itel)%PN1, & 
    !     myModelPSF(itel)%PM1, myModelPSF(itel)%BETA1, NORM1)
    !call polar_normalise_unity(SIZE( myModelPSF(itel)%NX2), PCOEFF2_TEMP, myModelPSF(itel)%PN2, &
    !     myModelPSF(itel)%PM2 , myModelPSF(itel)%BETA2, NORM2)             
    !PCOEFF1_TEMP = PCOEFF1_TEMP / (NORM1+NORM2)
    !PCOEFF2_TEMP = PCOEFF2_TEMP / (NORM1+NORM2)             
    !
    CCOEFF0_TEMP=matmul(TRANSPOSE(myModelPSF%MATRIX0), PCOEFF0_TEMP)
    CCOEFF1_TEMP=matmul(TRANSPOSE(myModelPSF%MATRIX1), PCOEFF1_TEMP)
    CCOEFF2_TEMP=matmul(TRANSPOSE(myModelPSF%MATRIX2), PCOEFF2_TEMP)
    COEFF0_TEMP = REALPART(CCOEFF0_TEMP)
    COEFF1_TEMP = REALPART(CCOEFF1_TEMP)
    COEFF2_TEMP = REALPART(CCOEFF2_TEMP)
    
    
!    CCOEFF1_TEMP = CMPLX( COEFF1_TEMP, 0)
!    CCOEFF2_TEMP = CMPLX( COEFF2_TEMP, 0)
!    
!          call slib_write2file(SIZE(myModelPSF%NX1), CCOEFF1_TEMP, &
!               myModelPSF%NX1, myModelPSF%NY1, &
!               real(myModelPSF%BETA1), &
!               SIZE(myModelPSF%NX2), CCOEFF2_TEMP, myModelPSF%NX2, &
!               myModelPSF%NY2, &
!               real(myModelPSF%BETA2), &
!               100, ISRC, IE, NP)
!          !if(NP>=10)stop
!          !stop
    
       
  end subroutine slib_get_xy



   subroutine findEN(E, myModelPSF, N)

    !use data_psf

    double precision :: E

    TYPE(MODELPSF) :: myModelPSF
   
    INTEGER N, FLAG
    !write(*,*) 
    N=0
    FLAG=0
    DO WHILE (FLAG==0 .AND. N<SIZE(myModelPSF%ENERGY) )
       N=N+1
       if(E<myModelPSF%ENERGY(N))FLAG=1
       !write(*,*) "H",E, N, myModelPSF%ENERGY(N), FLAG,SIZE(myModelPSF%ENERGY)

    ENDDO
    
    
    N=N-1
    IF(N<=0)N=1

    !******************
    ! this is only test
    !N=0
    ! this is only test
    !******************

    
  end subroutine findEN
  

 subroutine findXY(ARC_X, ARC_Y, myModelPSF, IXY)

    !use data_psf

    
    implicit none

    double precision :: ARC_X, ARC_Y
    
    TYPE(MODELPSF) :: myModelPSF

    integer IXY(:,:)

    integer IX, IY, M

    IX = ( ARC_X - myModelPSF%XMIN ) / myModelPSF%XBIN  + 1
    IY = ( ARC_Y - myModelPSF%YMIN ) / myModelPSF%YBIN  + 1
    M =  ( myModelPSF%YMAX - myModelPSF%YMIN ) / myModelPSF%YBIN + 1 


    !write(*,*) "IN",ARC_X, myModelPSF%XMIN, myModelPSF%XBIN, IX, M
    !write(*,*) "IN",ARC_Y, myModelPSF%YMIN, myModelPSF%YBIN, IY, M

    !  (1,2)    (2,2)         ^
    !                         |
    !  (1,1)    (2,1)   X-->  Y
    IXY(1,1) = (IX-1) * M + IY
    IXY(1,2) = (IX-1) * M + IY + 1
    IXY(2,1) = IX     * M + IY
    IXY(2,2) = IX     * M + IY + 1

    !******************
    ! this is only test
    !IXY=1
    ! this is only test
    !******************


  end subroutine findXY


  subroutine inter_linear( myModelPSF, IE, ENERGY, COEFF1, COEFF2)
    
    implicit none

    TYPE(MODELPSF), intent(in) :: myModelPSF
    double precision, intent(inout) :: COEFF1(:), COEFF2(:)
    integer, intent(in) :: IE    
    double precision, intent(in) ::  ENERGY

    double precision zd

    zd = ( ENERGY - myModelPSF%ENERGY(IE) ) & 
       / (myModelPSF%ENERGY(IE+1) - myModelPSF%ENERGY(IE) )

    !write(*,*) zd, ENERGY, myModelPSF%ENERGY(IE+1), myModelPSF%ENERGY(IE)

    COEFF1 = COEFF1  * (1.0 - zd) + COEFF2  * zd


  end subroutine inter_linear

  subroutine trilinear( myModelPSF, IXY, IE, ARC_X, ARC_Y, ENERGY, &
       COEFF0, COEFF1, COEFF2)

    !use data_psf

    implicit none
    
    TYPE(MODELPSF), intent(in) :: myModelPSF
    
    integer, intent(in) :: IXY(:,:)
    
    integer, intent(in) :: IE    
    
    double precision, intent(in) ::  ARC_X, ARC_Y, ENERGY
    
    double precision, intent(out) ::  COEFF0(:), COEFF1(:), COEFF2(:)
    
    integer IC 

    double precision xd, yd, zd
    double precision v_xo_yo_zo, v_x1_yo_zo
    double precision v_xo_y1_zo, v_x1_y1_zo
    double precision v_xo_yo_z1, v_x1_yo_z1
    double precision v_xo_y1_z1, v_x1_y1_z1
    
    double precision c00, c10, c01, c11
    double precision c0, c1, c
    
  !what I do is pointless. I should simply read all source 
  !without sorting them and the do interpolaion. 
  
  !write(*,*) "HERE:",IXY(1,1), IXY(2,2), SIZE(myModelPSF%XOFF)
  !write(*,*) "HERE:",IE, IE+1
 
  xd = ( ARC_X  - myModelPSF%XOFF( IXY(1,1) ) ) / & 
       ( myModelPSF%XOFF( IXY(2,2) ) - myModelPSF%XOFF( IXY(1,1) ) )
  yd = (  ARC_Y - myModelPSF%YOFF( IXY(1,1) ) ) / &
       ( myModelPSF%YOFF( IXY(2,2) ) - myModelPSF%YOFF( IXY(1,1) ) )
  zd = ( ENERGY - myModelPSF%ENERGY(IE) ) & 
       / (myModelPSF%ENERGY(IE+1) - myModelPSF%ENERGY(IE) )
  
  !write(*,*)
  !write(*,*) "TRI START"
  !write(*,*) myModelPSF%XOFF( IXY(1,1) ),  myModelPSF%XOFF( IXY(2,2) )
  !write(*,*) ARC_X, ARC_Y
  !write(*,*) myModelPSF%ENERGY(IE), myModelPSF%ENERGY(IE+1),  ENERGY
  !write(*,*) "XD, YD, ZD:",xd, yd, zd


  do IC=1,SIZE( myModelPSF%NX0 )

     v_xo_yo_zo=myModelPSF%C0(IC, IXY(1,1), IE)
     v_x1_yo_zo=myModelPSF%C0(IC, IXY(2,1), IE)

     !write(*,*) "C0: v_xo_yo_zo=",IC, IXY(1,1), IE, myModelPSF%C0(IC, IXY(1,1), IE)
     !write(*,*) "C0: v_x1_yo_zo=",IC, IXY(2,1), IE, myModelPSF%C0(IC, IXY(2,1), IE)
     
     v_xo_y1_zo=myModelPSF%C0(IC, IXY(1,2), IE)
     v_x1_y1_zo=myModelPSF%C0(IC, IXY(2,2), IE)
     
     !write(*,*) "C0: v_xo_y1_zo=",IC, IXY(1,2), IE, myModelPSF%C0(IC, IXY(1,2), IE)
     !write(*,*) "C0: v_x1_y1_zo=",IC, IXY(2,2), IE, myModelPSF%C0(IC, IXY(2,2), IE)
     
     v_xo_yo_z1=myModelPSF%C0(IC, IXY(1,1), IE+1)
     v_x1_yo_z1=myModelPSF%C0(IC, IXY(2,1), IE+1)
           
     !write(*,*) "C0: v_xo_yo_z1=",IC,  IXY(1,1), myModelPSF%C0(IC, IXY(1,1), IE+1)
     !write(*,*) "C0: v_x1_yo_z1=",IC, IXY(2,1), myModelPSF%C0(IC, IXY(2,1), IE+1)

     v_xo_y1_z1=myModelPSF%C0(IC, IXY(1,2), IE+1)
     v_x1_y1_z1=myModelPSF%C0(IC, IXY(2,2), IE+1)
  
     !write(*,*) "C1: v_xo_y1_z1=",IC, IXY(1,2), myModelPSF%C0(IC, IXY(1,2), IE+1)
     !write(*,*) "C1: v_x1_y1_z1=",IC, IXY(2,2), myModelPSF%C0(IC, IXY(2,2), IE+1)
          
     !  write(22,*) "X0, Y0, Z0:", cr, hr1, hr2, ":",  v_xo_yo_zo, " X1, Y0, Z0:", cr_t, hr1, hr2, ":", v_x1_yo_zo
     !  write(22,*) "X0, Y1, Z0:", cr, hr1_t, hr2, ":", v_xo_y1_zo," X1, Y1, Z0:", cr_t, hr1_t, hr2, ":", v_x1_y1_zo
     !  write(22,*) "X0, Y0, Z1:", cr, hr1, hr2_t, ":", v_xo_yo_z1," X1, Y0, Z1:", cr_t, hr1, hr2_t,":", v_x1_yo_z1
     !  write(22,*) "X0, Y1, Z1:", cr, hr1_t, hr2_t, ":", v_xo_y1_z1, " X1, Y1, Z1:", cr_t, hr1_t, hr2_tn=, ":",v_x1_y1_z1
     
     
     c00= v_xo_yo_zo * (1.0 - xd) + v_x1_yo_zo * xd
     c10= v_xo_y1_zo * (1.0 - xd) + v_x1_y1_zo * xd
     c01= v_xo_yo_z1 * (1.0 - xd) + v_x1_yo_z1 * xd
     c11= v_xo_y1_z1 * (1.0 - xd) + v_x1_y1_z1 * xd
     
     !  write(22,*) "C00:",c00
     !  write(22,*) "C10:",c10
     !  write(22,*) "C01:",c01
     !  write(22,*) "C11:",c11
     
     c0 = c00 * (1.0 - yd) + c10 * yd
     c1 = c01 * (1.0 - yd) + c11 * yd
     c  = c0  * (1.0 - zd) + c1  * zd
     
     !write(*,*) "C=",IC, c
     
     COEFF0(IC)=c

  enddo



  do IC=1,SIZE( myModelPSF%NX1 )

     v_xo_yo_zo=myModelPSF%C1(IC, IXY(1,1), IE)
     v_x1_yo_zo=myModelPSF%C1(IC, IXY(2,1), IE)

     !if(IC==1)then
     !   write(*,*) "C1: v_xo_yo_zo=",IC, IXY(1,1), IE, myModelPSF%C1(IC, IXY(1,1), IE)
     !   write(*,*) "C1: v_x1_yo_zo=",IC, IXY(2,1), IE, myModelPSF%C1(IC, IXY(2,1), IE)
     !endif
     
     v_xo_y1_zo=myModelPSF%C1(IC, IXY(1,2), IE)
     v_x1_y1_zo=myModelPSF%C1(IC, IXY(2,2), IE)
     
     !if(IC==1)then
     !   write(*,*) "C1: v_xo_y1_zo=",IC, IXY(1,2), IE, myModelPSF%C1(IC, IXY(1,2), IE)
     !   write(*,*) "C1: v_x1_y1_zo=",IC, IXY(2,2), IE, myModelPSF%C1(IC, IXY(2,2), IE)
     !endif
     
     v_xo_yo_z1=myModelPSF%C1(IC, IXY(1,1), IE+1)
     v_x1_yo_z1=myModelPSF%C1(IC, IXY(2,1), IE+1)
           
     !write(*,*) "C1: v_xo_yo_z1=",IC, IXY(1,1), myModelPSF%C1(IC, IXY(1,1), IE+1)
     !write(*,*) "C1: v_x1_yo_z1=",IC, IXY(2,1), myModelPSF%C1(IC, IXY(2,1), IE+1)

     v_xo_y1_z1=myModelPSF%C1(IC, IXY(1,2), IE+1)
     v_x1_y1_z1=myModelPSF%C1(IC, IXY(2,2), IE+1)
  
     !write(*,*) "C1: v_xo_y1_z1=",IC, IXY(1,2), myModelPSF%C1(IC, IXY(1,2), IE+1)
     !write(*,*) "C1: v_x1_y1_z1=",IC, IXY(2,2),  myModelPSF%C1(IC, IXY(2,2), IE+1)
          
     !  write(22,*) "X0, Y0, Z0:", cr, hr1, hr2, ":",  v_xo_yo_zo, " X1, Y0, Z0:", cr_t, hr1, hr2, ":", v_x1_yo_zo
     !  write(22,*) "X0, Y1, Z0:", cr, hr1_t, hr2, ":", v_xo_y1_zo," X1, Y1, Z0:", cr_t, hr1_t, hr2, ":", v_x1_y1_zo
     !  write(22,*) "X0, Y0, Z1:", cr, hr1, hr2_t, ":", v_xo_yo_z1," X1, Y0, Z1:", cr_t, hr1, hr2_t,":", v_x1_yo_z1
     !  write(22,*) "X0, Y1, Z1:", cr, hr1_t, hr2_t, ":", v_xo_y1_z1, " X1, Y1, Z1:", cr_t, hr1_t, hr2_tn=, ":",v_x1_y1_z1
     
     
     c00= v_xo_yo_zo * (1.0 - xd) + v_x1_yo_zo * xd
     c10= v_xo_y1_zo * (1.0 - xd) + v_x1_y1_zo * xd
     c01= v_xo_yo_z1 * (1.0 - xd) + v_x1_yo_z1 * xd
     c11= v_xo_y1_z1 * (1.0 - xd) + v_x1_y1_z1 * xd
     
     !  write(22,*) "C00:",c00
     !  write(22,*) "C10:",c10
     !  write(22,*) "C01:",c01
     !  write(22,*) "C11:",c11
     
     c0 = c00 * (1.0 - yd) + c10 * yd
     c1 = c01 * (1.0 - yd) + c11 * yd
     c  = c0  * (1.0 - zd) + c1  * zd
     
     !write(*,*) "C=",IC, c
     
     COEFF1(IC)=c

  enddo



  do IC=1,SIZE( myModelPSF%NX2 )

     v_xo_yo_zo=myModelPSF%C2(IC, IXY(1,1), IE)
     v_x1_yo_zo=myModelPSF%C2(IC, IXY(2,1), IE)

     !if(IC==1)then
     !   write(*,*) "C2: v_xo_yo_zo=",IC, IXY(1,1), IE, myModelPSF%C2(IC, IXY(1,1), IE)
     !   write(*,*) "C2: v_x1_yo_zo=",IC, IXY(2,1), IE, myModelPSF%C2(IC, IXY(2,1), IE)
     !endif
     
     v_xo_y1_zo=myModelPSF%C2(IC, IXY(1,2), IE)
     v_x1_y1_zo=myModelPSF%C2(IC, IXY(2,2), IE)
     
     !if(IC==1)then
     !   write(*,*) "C2: v_xo_y1_zo=",IC, IXY(1,2), IE, myModelPSF%C2(IC, IXY(1,2), IE)
     !   write(*,*) "C2: v_x1_y1_zo=",IC, IXY(2,2), IE, myModelPSF%C2(IC, IXY(2,2), IE)
     !endif
     
     v_xo_yo_z1=myModelPSF%C2(IC, IXY(1,1), IE+1)
     v_x1_yo_z1=myModelPSF%C2(IC, IXY(2,1), IE+1)
           
     !write(*,*) "C2: v_xo_yo_z1=",IC, IXY(1,1), myModelPSF%C2(IC, IXY(1,1), IE+1)
     !write(*,*) "C2: v_x1_yo_z1=",IC, IXY(2,1), myModelPSF%C2(IC, IXY(2,1), IE+1)

     v_xo_y1_z1=myModelPSF%C2(IC, IXY(1,2), IE+1)
     v_x1_y1_z1=myModelPSF%C2(IC, IXY(2,2), IE+1)
  
     !write(*,*) "C2: v_xo_y1_z1=",IC, IXY(1,2), myModelPSF%C2(IC, IXY(1,2), IE+1)
     !write(*,*) "C2: v_x1_y1_z1=",IC, IXY(2,2), myModelPSF%C2(IC, IXY(2,2), IE+1)
          
     !  write(22,*) "X0, Y0, Z0:", cr, hr1, hr2, ":",  v_xo_yo_zo, " X1, Y0, Z0:", cr_t, hr1, hr2, ":", v_x1_yo_zo
     !  write(22,*) "X0, Y1, Z0:", cr, hr1_t, hr2, ":", v_xo_y1_zo," X1, Y1, Z0:", cr_t, hr1_t, hr2, ":", v_x1_y1_zo
     !  write(22,*) "X0, Y0, Z1:", cr, hr1, hr2_t, ":", v_xo_yo_z1," X1, Y0, Z1:", cr_t, hr1, hr2_t,":", v_x1_yo_z1
     !  write(22,*) "X0, Y1, Z1:", cr, hr1_t, hr2_t, ":", v_xo_y1_z1, " X1, Y1, Z1:", cr_t, hr1_t, hr2_tn=, ":",v_x1_y1_z1
     
     
     c00= v_xo_yo_zo * (1.0 - xd) + v_x1_yo_zo * xd
     c10= v_xo_y1_zo * (1.0 - xd) + v_x1_y1_zo * xd
     c01= v_xo_yo_z1 * (1.0 - xd) + v_x1_yo_z1 * xd
     c11= v_xo_y1_z1 * (1.0 - xd) + v_x1_y1_z1 * xd
     
     !  write(22,*) "C00:",c00
     !  write(22,*) "C10:",c10
     !  write(22,*) "C01:",c01
     !  write(22,*) "C11:",c11
     
     c0 = c00 * (1.0 - yd) + c10 * yd
     c1 = c01 * (1.0 - yd) + c11 * yd
     c  = c0  * (1.0 - zd) + c1  * zd
     
     !write(*,*) "C=",c
     
     COEFF2(IC)=c

  enddo
  
end subroutine trilinear



  subroutine polar_rotate(N_COEFF, P_COEFF, LOOKUP_N, LOOKUP_M, THETA_INPUT)


    integer, INTENT(IN) :: N_COEFF
    
    complex P_COEFF(N_COEFF)
    
    integer, INTENT(IN), DIMENSION(N_COEFF) :: LOOKUP_N, LOOKUP_M
    
    double precision THETA_INPUT, THETA
    
    double precision PHI, AMPL, X, Y

    INTEGER i, M, N
    
    !write(*,*) THETA, N_COEFF, SIZE( LOOKUP_N),SIZE( LOOKUP_M)
    !write(*,*) P_COEFF
    !write(*,*) myDEG2RAD * THETA
    !write(*,*) LOOKUP_M
     
    THETA = THETA_INPUT * myDEG2RAD

    !write(*,*) THETA
    
    do i = 1, N_COEFF
       M=LOOKUP_M(I)
       N=LOOKUP_N(I)
       
       !write(*,*) N, M, P_COEFF(i)

       X=REALPART(P_COEFF(i))
       Y=IMAGPART(P_COEFF(i))
       
       !write(*,*) "X, Y=", X, Y

       PHI=atan2(Y, X)
       
       !write(*,*) "PHI=", PHI

       PHI=PHI + M * THETA 
       
       AMPL = sqrt( X * X + Y * Y)
       
       !write(*,*) "NEW AMPL, PHI=", AMPL, PHI
       
       
       P_COEFF(i) = CMPLX (AMPL * COS(PHI) , AMPL * SIN(PHI))
       
       !write(*,*) "NEW COMPLEX:", i, P_COEFF(i)
       !WRITE(*,*) 
    enddo
    
  end subroutine polar_rotate






  subroutine polar_normalise_unity(N_COEFF, P_COEFF, LOOKUP_N, LOOKUP_M, BETA, NORM)


    integer, INTENT(IN) :: N_COEFF
    
    complex, INTENT(IN) :: P_COEFF(N_COEFF)
    
    integer, INTENT(IN), DIMENSION(N_COEFF) :: LOOKUP_N, LOOKUP_M
    
    double precision, INTENT(IN) :: BETA

    double precision, INTENT(OUT) :: NORM

    double precision SUM, X, Y
    

    INTEGER i, M, N
    
    !write(*,*) THETA, N_COEFF, SIZE( LOOKUP_N),SIZE( LOOKUP_M)
    !write(*,*) P_COEFF
    !write(*,*) myDEG2RAD * THETA
    !write(*,*) LOOKUP_M
     
    SUM=0.0

    do i = 1, N_COEFF
       M=LOOKUP_M(I)
       N=LOOKUP_N(I)
       
       !write(*,*) N, M, P_COEFF(i)

       X=REALPART(P_COEFF(i))
       Y=IMAGPART(P_COEFF(i))
       !write(*,*) "N,M, X, Y=", N,M,X,Y 
       if(M.eq.0.and.MOD(N,2).eq.0)then
          SUM = SUM + X
          !write(*,*) "X, Y, SUM=", X, Y, SUM
       endif       
       
    enddo

    NORM = BETA * SQRT (myPI * 4.) * SUM
    
   !write(*,*) "NORM, BETA", NORM, BETA

  end subroutine polar_normalise_unity




  subroutine slib_fluxEEF(myStackPSF, myModelPSF, ISRC, M, EEF, RADIUS)

    !include for testing
    !use slib_cimage

    implicit none

    TYPE(STACKPSF), intent(in) :: myStackPSF
    TYPE(MODELPSF), intent(in) :: myModelPSF
    INTEGER, intent(in) :: ISRC
    INTEGER, intent(in) :: M
    DOUBLE PRECISION, DIMENSION(M),  intent(in) :: EEF
    DOUBLE PRECISION, DIMENSION(M),  intent(out) :: RADIUS

    complex P_COEFF0(SIZE( myModelPSF%NX0)), &
         P_COEFF1(SIZE( myModelPSF%NX1)),  P_COEFF2(SIZE( myModelPSF%NX2)) 
    complex C_COEFF0(SIZE( myModelPSF%NX0)), &
         C_COEFF1(SIZE( myModelPSF%NX1)),  C_COEFF2(SIZE( myModelPSF%NX2)) 
    DOUBLE PRECISION, DIMENSION(100) ::  R, RES0, RES1, RES2, FLUXR
    DOUBLE PRECISION SUBSAMPLE, BETA
    DOUBLE PRECISION NORM0, NORM1, NORM2, NORM
    INTEGER I,J
    INTEGER, parameter :: NR=100
    logical FLAG

    ! this will be R in units of RAW_PIX_MICRON
    ! I will convert R [units  myCPSF%IMG_PIXEL_MM]
    ! to Rprime and then estimate flux at this new radius
    DOUBLE PRECISION, DIMENSION(NR) :: RPRIME

    
    !include for testing
    !INTEGER IE, NP, I
    !DOUBLE PRECISION ANGLE
    !ANGLE = 0.

    do I=1,NR
       R(I)=DBLE(I)       
    enddo

    !write(*,*) "R=", R
    SUBSAMPLE = RAW_PIX_MICRON / myModelPSF%IMG_PIXEL_MM
    RPRIME = R * myModelPSF%IMG_PIXEL_MM / RAW_PIX_MICRON


    C_COEFF0 = CMPLX( myStackPSF%C0(:, ISRC), 0)
    C_COEFF1 = CMPLX( myStackPSF%C1(:, ISRC), 0)
    C_COEFF2 = CMPLX( myStackPSF%C2(:, ISRC), 0)
 

    RES0=0.0
    P_COEFF0=matmul(TRANSPOSE(myModelPSF%MATRIX_INV0), C_COEFF0)        
    BETA  = myModelPSF%BETA0 !* SUBSAMPLE
    call slib_calc_fluxR(SIZE(myModelPSF%NX0), BETA, P_COEFF0, myModelPSF%PN0, myModelPSF%PM0,  NR, RPRIME, RES0)
    call polar_normalise_unity(SIZE(myModelPSF%NX0), P_COEFF0, myModelPSF%PN0, myModelPSF%PM0, BETA, NORM0)

    RES1=0.0
    P_COEFF1=matmul(TRANSPOSE(myModelPSF%MATRIX_INV1), C_COEFF1)        
    BETA  = myModelPSF%BETA1 !* SUBSAMPLE
    call slib_calc_fluxR(SIZE(myModelPSF%NX1), BETA, P_COEFF1, myModelPSF%PN1, myModelPSF%PM1,  NR, RPRIME, RES1)
    call polar_normalise_unity(SIZE(myModelPSF%NX1), P_COEFF1, myModelPSF%PN1, myModelPSF%PM1, BETA, NORM1)

    !include for testing
    !call polar_rotate(SIZE(myStackPSF%NX1), P_COEFF1, myStackPSF%PN1, & 
    !              myStackPSF%PM1, ANGLE)  
    RES2=0.0
    P_COEFF2=matmul(TRANSPOSE(myModelPSF%MATRIX_INV2), C_COEFF2)    
    BETA  = myModelPSF%BETA2 !* SUBSAMPLE
    call  slib_calc_fluxR( SIZE(myModelPSF%NX2), BETA, P_COEFF2, myModelPSF%PN2, myModelPSF%PM2,  NR, RPRIME, RES2)
    call polar_normalise_unity(SIZE(myModelPSF%NX2), P_COEFF2, myModelPSF%PN2, myModelPSF%PM2, BETA, NORM2)


    FLUXR = RES0 + RES1 + RES2
    NORM  = NORM0 + NORM1 + NORM2

    FLUXR = FLUXR / NORM ! RETURN EEF

    !do J = 1, NR
    !  write(40,*) R(J), FLUXR(J), ISRC
    !enddo
    !write(*,*) FLUXR, "TOTAL=",NORM

    !include for testing
    RADIUS=0.0
    DO I=1,M

       J=1
       FLAG=.TRUE.
       DO WHILE (J<=NR .and. FLAG)
          if(  FLUXR(J) >= EEF(I)) then
             FLAG=.FALSE.

             if(J > 1)then
                
                !write(*,*) "LOOP:",I, EEF(I), J,  FLUXR(J), SIZE(R), SIZE(FLUXR)

                RADIUS(I) = R(J-1) + &
                     ( EEF(I) - FLUXR(J-1)) * &
                     ( R(J-1) - R(J) ) / ( FLUXR(J-1) - FLUXR(J) )

                !write(*,*) "END:",  FLUXR(J-1), EEF(I) ,  FLUXR(J)
                !write(*,*) "END:",  R(J-1), RADIUS(I) ,  R(J)

             ELSE
                RADIUS(I) =  R(1)

             ENDIF
          ENDIF
          
          J = J +1

       ENDDO
    ENDDO
    !write(*,*) RADIUS
    !write(*,*) EEF
    !   write(*,*) R(I), FLUXR(I), FLUXR(I)/NORM
    !   write(20,*) R(I), FLUXR(I), FLUXR(I)/NORM
    !ENDDO
    !stop

  end subroutine slib_fluxEEF



  subroutine slib_fluxR(myStackPSF, myCPSF, ISRC, M, R, FLUXR)

    !include for testing
    !use slib_cimage

    implicit none

    TYPE(STACKPSF), intent(in) :: myStackPSF
    TYPE(CPSF), intent(in) :: myCPSF
    INTEGER, intent(in) :: ISRC
    INTEGER, intent(in) :: M
    DOUBLE PRECISION, DIMENSION(M),  intent(in) :: R
    DOUBLE PRECISION, DIMENSION(M),  intent(out) :: FLUXR

    ! this will be R in units of RAW_PIX_MICRON
    ! I will convert R [units  myCPSF%IMG_PIXEL_MM]
    ! to Rprime and then estimate flux at this new radius
    DOUBLE PRECISION, DIMENSION(M) :: RPRIME
    
    complex  P_COEFF0(SIZE( myCPSF%NX0)), &
         P_COEFF1(SIZE( myCPSF%NX1)),  P_COEFF2(SIZE( myCPSF%NX2)) 
    complex C_COEFF0(SIZE( myCPSF%NX0)), &
         C_COEFF1(SIZE( myCPSF%NX1)),  C_COEFF2(SIZE( myCPSF%NX2)) 
    DOUBLE PRECISION, DIMENSION(M) ::  RES0, RES1, RES2
    DOUBLE PRECISION SUBSAMPLE, BETA
    DOUBLE PRECISION NORM0, NORM1, NORM2, NORM

    !include for testing
    !INTEGER IE, NP, I
    !DOUBLE PRECISION ANGLE
    !ANGLE = 0.

    SUBSAMPLE = RAW_PIX_MICRON / myCPSF%IMG_PIXEL_MM
    RPRIME = R * myCPSF%IMG_PIXEL_MM / RAW_PIX_MICRON

    C_COEFF0 = CMPLX( myStackPSF%C0(:, ISRC), 0)
    C_COEFF1 = CMPLX( myStackPSF%C1(:, ISRC), 0)
    C_COEFF2 = CMPLX( myStackPSF%C2(:, ISRC), 0)
 
    !write(*,*) "SOURCE ", ISRC
    !write(*,*) "BETA0, SIZE NX0", myCPSF%BETA0, SIZE(myStackPSF%NX0)
    !write(*,*) "BETA1, SIZE NX1", myCPSF%BETA1, SIZE(myStackPSF%NX1)
    !write(*,*) "BETA2, SIZE NX2", myCPSF%BETA2, SIZE(myStackPSF%NX2)

    !WRITE(*,*) myStackPSF%C1(:, ISRC)
    !WRITE(*,*) myStackPSF%C2(:, ISRC)
    !WRITE(*,*) C_COEFF1
    !WRITE(*,*) C_COEFF2

    RES0=0.0
    P_COEFF0=matmul(TRANSPOSE(myCPSF%MATRIX_INV0), C_COEFF0)        
    BETA  = myCPSF%BETA0 !* SUBSAMPLE
    call slib_calc_fluxR(SIZE(myCPSF%NX0), BETA, P_COEFF0, myCPSF%PN0, myCPSF%PM0,  M, RPRIME, RES0)
    call polar_normalise_unity(SIZE(myCPSF%NX0), P_COEFF0, myCPSF%PN0, myCPSF%PM0, BETA, NORM0)

    RES1=0.0
    P_COEFF1=matmul(TRANSPOSE(myCPSF%MATRIX_INV1), C_COEFF1)        
    BETA  = myCPSF%BETA1 !* SUBSAMPLE
    call slib_calc_fluxR(SIZE(myCPSF%NX1), BETA, P_COEFF1, myCPSF%PN1, myCPSF%PM1,  M, RPRIME, RES1)
    call polar_normalise_unity(SIZE(myCPSF%NX1), P_COEFF1, myCPSF%PN1, myCPSF%PM1, BETA, NORM1)

    !include for testing
    !call polar_rotate(SIZE(myStackPSF%NX1), P_COEFF1, myStackPSF%PN1, & 
    !              myStackPSF%PM1, ANGLE)  
    RES2=0.0
    P_COEFF2=matmul(TRANSPOSE(myCPSF%MATRIX_INV2), C_COEFF2)    
    BETA  = myCPSF%BETA2 !* SUBSAMPLE
    call  slib_calc_fluxR( SIZE(myCPSF%NX2), BETA, P_COEFF2, myCPSF%PN2, myCPSF%PM2,  M, RPRIME, RES2)
    call polar_normalise_unity(SIZE(myCPSF%NX2), P_COEFF2, myCPSF%PN2, myCPSF%PM2, BETA, NORM2)

    ! including for testing
    !call polar_rotate(SIZE(myCPSF%NX2), P_COEFF2, myCPSF%PN2, & 
    !              myCPSF%PM2, ANGLE)  

    !C_COEFF1=matmul(TRANSPOSE(myCPSF%MATRIX1), P_COEFF1)
    !C_COEFF2=matmul(TRANSPOSE(myCPSF%MATRIX2), P_COEFF2)
    !C_COEFF1 = REALPART( C_COEFF1 )
    !C_COEFF2 = REALPART( C_COEFF2 )

    !IE=1
    !NP=1
    !call slib_write2file(SIZE(myCPSF%NX1), C_COEFF1, &
    !     myCPSF%NX1, myCPSF%NY1, &
    !     real(myCPSF%BETA1 * SUBSAMPLE), &
    !     SIZE(myCPSF%NX2), C_COEFF2, myCPSF%NX2, &
    !     myCPSF%NY2, &
    !     real(myCPSF%BETA2 * SUBSAMPLE), &
    !     100, ISRC, IE, NP)


    !write(*,*) RES1
    !write(*,*) RES2

    FLUXR = RES0 + RES1 + RES2
    NORM  = NORM0 + NORM1 + NORM2

    !FLUXR = FLUXR / NORM ! RETURN EEF

    !write(*,*) FLUXR, "TOTAL=",NORM

    !include for testing
    !DO I=1,SIZE(RES1)
    !   write(*,*) R(I), FLUXR(I), FLUXR(I)/NORM
    !   write(20,*) R(I), FLUXR(I), FLUXR(I)/NORM
    !ENDDO
    !stop

  end subroutine slib_fluxR


  subroutine slib_calc_fluxR(N_COEFF, BETA, P_COEFF, LOOKUP_N, LOOKUP_M,  K , R, FLUXR )

    use slib_basic_math


    implicit none

    integer, INTENT(IN) :: N_COEFF
    complex, INTENT(IN) :: P_COEFF(N_COEFF)
    integer, INTENT(IN), DIMENSION(N_COEFF) :: LOOKUP_N, LOOKUP_M
    double precision, INTENT(IN) :: BETA    
    INTEGER, intent(in) :: K
    DOUBLE PRECISION, DIMENSION(K),  intent(in) :: R
    DOUBLE PRECISION, DIMENSION(K),  intent(out) :: FLUXR



    DOUBLE PRECISION, ALLOCATABLE :: VAL(:,:), SUMVAL(:,:), SUM_LAGUERRE(:,:)
    DOUBLE PRECISION, DIMENSION(K) :: R2_OVER_2BETA2, R2_OVER_BETA2, EXP_R2_OVER_2BETA2, &
         IN, SUM1, SUM2
    DOUBLE PRECISION X, Y

    INTEGER I, J, NMAX, M, N


    R2_OVER_2BETA2 = R**2 / 2.0 / BETA**2
    R2_OVER_BETA2 = R**2 / BETA**2
    EXP_R2_OVER_2BETA2 = EXP(-R2_OVER_2BETA2)

    ! HERE I FIND THE MAX EVEN POLAR ORDER N AND DEVIDE IT BY 2 
    ! J IS THE MAX OF N POLAR ORDER
    if( MOD(N_COEFF,2) .eq. 0) then
       NMAX=N_COEFF/2 ! I should make sure this INTEGER
    else
       NMAX=(N_COEFF-1)/2 ! I should make sure this INTEGER
    endif

    ALLOCATE ( VAL (K, 0:NMAX) ) ! REMEMBER TO DE-ALLOCATE ON EXIT
    ALLOCATE ( SUMVAL (K, 0:NMAX) ) ! REMEMBER TO DE-ALLOCATE ON EXIT
    ALLOCATE ( SUM_LAGUERRE (K, 0:NMAX) ) ! REMEMBER TO DE-ALLOCATE ON EXIT


    call l_polynomial ( K, NMAX, R2_OVER_BETA2, VAL )

    ! I NEED TO ESTIMATE SUM(I=0..NMAX)[ (-1)**I * LAGUERRE_I(X) ] 
    ! FIRST I ESTIMATE ARRAY (-1)**I * LAGUERRE_I(X) TO SUM UP LATER
    DO J = 0, NMAX
       SUMVAL(:,J) = VAL(:,J) * (-1)**J
    ENDDO

    ! HERE I ESSENTIALLY PRE-ESTIMATE SUM(I=0..NMAX)[ (-1)**I * LAGUERRE_I(X) ] 
    ! FOR ALL Xs AND ALL Is
    do J = 0, NMAX
       do I = 1, K
          SUM_LAGUERRE(I, J) = SUM(SUMVAL(I,0:J), 1)
       enddo
    enddo

    SUM1 = 0 
    do I = 1,N_COEFF

       M=LOOKUP_M(I)
       N=LOOKUP_N(I)

       if( M.eq.0 .AND. MOD(N,2).eq.0 )then
          
          J = N/2

          IN = 1.0 - EXP_R2_OVER_2BETA2 * ( 2.0 *  SUM_LAGUERRE(:, J)  - (-1)**J * VAL(:,J) )

          X=REALPART(P_COEFF(i))
          !Y=IMAGPART(P_COEFF(i))

          SUM1 = SUM1 + IN * P_COEFF(I)

       endif

    enddo

    DEALLOCATE ( VAL, SUMVAL, SUM_LAGUERRE)

    
    FLUXR = BETA * SQRT (myPI * 4.) * SUM1 

  end subroutine slib_calc_fluxR







end module slib_calc_space
