! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module slib_data_evt
  
  use slib_constants
  
  implicit none
  
  private

  public :: slib_deallocate_event, slib_evttool_2_myevent, slib_sort_heal
   


contains
  
  subroutine slib_deallocate_event(myEVENTFILE)

    implicit none
    
    TYPE(EVENTFILE) :: myEVENTFILE

    integer ii

    IF (ALLOCATED (myEVENTFILE%EVENT%XD)) DEALLOCATE (myEVENTFILE%EVENT%XD)
    IF (ALLOCATED (myEVENTFILE%EVENT%YD)) DEALLOCATE (myEVENTFILE%EVENT%YD)
    IF (ALLOCATED (myEVENTFILE%EVENT%ZD)) DEALLOCATE (myEVENTFILE%EVENT%ZD)
 
    IF (ALLOCATED (myEVENTFILE%EVENT%RA)) DEALLOCATE (myEVENTFILE%EVENT%RA)
    IF (ALLOCATED (myEVENTFILE%EVENT%DEC)) DEALLOCATE (myEVENTFILE%EVENT%DEC)
    IF (ALLOCATED (myEVENTFILE%EVENT%RAWX)) DEALLOCATE (myEVENTFILE%EVENT%RAWX)
    IF (ALLOCATED (myEVENTFILE%EVENT%RAWY)) DEALLOCATE (myEVENTFILE%EVENT%RAWY)
    IF (ALLOCATED (myEVENTFILE%EVENT%ENERGY)) DEALLOCATE (myEVENTFILE%EVENT%ENERGY)
    IF (ALLOCATED (myEVENTFILE%EVENT%TIME)) DEALLOCATE (myEVENTFILE%EVENT%TIME)
    IF (ALLOCATED (myEVENTFILE%EVENT%ROLL)) DEALLOCATE (myEVENTFILE%EVENT%ROLL)
    IF (ALLOCATED (myEVENTFILE%EVENT%TIME)) DEALLOCATE (myEVENTFILE%EVENT%TIME)
    IF (ALLOCATED (myEVENTFILE%EVENT%EV2SRC)) DEALLOCATE (myEVENTFILE%EVENT%EV2SRC)
    IF (ALLOCATED (myEVENTFILE%EVENT%CCDNR)) DEALLOCATE (myEVENTFILE%EVENT%CCDNR)


    IF (ALLOCATED (myEVENTFILE%EVENT%IPIX)) DEALLOCATE (myEVENTFILE%EVENT%IPIX)
    IF (ALLOCATED (myEVENTFILE%EVENT%INDEX1)) DEALLOCATE (myEVENTFILE%EVENT%INDEX1)
    IF (ALLOCATED (myEVENTFILE%EVENT%INDEX2)) DEALLOCATE (myEVENTFILE%EVENT%INDEX2)


    IF (ALLOCATED (myEVENTFILE%WHICH_CCDNR)) DEALLOCATE  (myEVENTFILE%WHICH_CCDNR)

    do ii = 1 , myEVENTFILE%NUM_ATT  
       
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%RA)) DEALLOCATE (myEVENTFILE%ATT(ii)%RA)
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%DEC)) DEALLOCATE (myEVENTFILE%ATT(ii)%DEC)
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%ROLL)) DEALLOCATE (myEVENTFILE%ATT(ii)%ROLL)
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%TIME)) DEALLOCATE (myEVENTFILE%ATT(ii)%TIME)
    enddo
    
     IF (ALLOCATED ( myEVENTFILE%ATT ) )  DEALLOCATE (myEVENTFILE%ATT) 

     myEVENTFILE%EVENT%HEALSORT = .False.
     
   end subroutine slib_deallocate_event



  subroutine slib_evttool_2_myevent(myEVENTFILE, evttoolEVENT)

    use slib_basic_calcs   
    use evtool_mod
    use evtool_eventfile_mod
    use slib_heal
 
    
    implicit none
    
    type(tEventFile) evttoolEVENT
    TYPE(EVENTFILE) myEVENTFILE

    integer nrows_ev, ii, nrows, max_nrows
    double precision ra, dec, xd, yd, zd
    integer var
    
    ! allocate EVENT
    nrows_ev = evttoolEVENT%events%nelements

    myEVENTFILE%EVENT%SIZE = nrows_ev

    allocate  ( myEVENTFILE%EVENT%XD(nrows_ev), myEVENTFILE%EVENT%YD(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ZD(nrows_ev) )    
    allocate  ( myEVENTFILE%EVENT%RA(nrows_ev), myEVENTFILE%EVENT%DEC(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%RAWX(nrows_ev), myEVENTFILE%EVENT%RAWY(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ENERGY(nrows_ev), myEVENTFILE%EVENT%TIME(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ROLL(nrows_ev), myEVENTFILE%EVENT%CCDNR(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%EV2SRC(nrows_ev) )

    ! this estimates the number of pixels, pixels start from 0
    call NumPixels(myEVENTFILE%EVENT%nside, var)
    allocate  ( myEVENTFILE%EVENT%INDEX1(0:var) )
    allocate  ( myEVENTFILE%EVENT%INDEX2(0:var) )

    allocate  ( myEVENTFILE%EVENT%IPIX(nrows_ev) )

    
    myEVENTFILE%EVENT%RA = evttoolEVENT%events%ra
    myEVENTFILE%EVENT%DEC = evttoolEVENT%events%dec
    myEVENTFILE%EVENT%RAWX = evttoolEVENT%events%rawx
    myEVENTFILE%EVENT%RAWY = evttoolEVENT%events%rawy
    myEVENTFILE%EVENT%ENERGY = evttoolEVENT%events%energy
    myEVENTFILE%EVENT%TIME = evttoolEVENT%events%time
    myEVENTFILE%EVENT%CCDNR = evttoolEVENT%events%ccdnr


    myEVENTFILE%EVENT%ROLL=0.D0
    myEVENTFILE%EVENT%EV2SRC=0

    ! this is healpix - sorting pixels by ra and dec in healpix
    call SkyToPixTable(myEVENTFILE%EVENT%RA, myEVENTFILE%EVENT%DEC, myEVENTFILE%EVENT%nside, myEVENTFILE%EVENT%IPIX)


    !write(*,*) "IN:",nrows_ev, size(myEVENTFILE%EVENT%XD)
    
    do ii=1, nrows_ev

       
       ra= myEVENTFILE%EVENT%RA(ii) / myRAD2DEG
       dec= myEVENTFILE%EVENT%DEC(ii) / myRAD2DEG


       call  slib_src_pos2xyz ( ra, dec, xd, yd, zd)
       
       myEVENTFILE%EVENT%XD(ii)= xd
       myEVENTFILE%EVENT%YD(ii)= yd
       myEVENTFILE%EVENT%ZD(ii)= zd

    enddo
    ! allocate event end


    ! ATTITUDE FILE
    ! first allocate the %ATT array of evt with the number of 
    ! atttitude extensions / number of active instruments   
    myEVENTFILE%NUM_ATT = evttoolEVENT%ninst
    allocate  ( myEVENTFILE%ATT(myEVENTFILE%NUM_ATT) )

    ! this step assigns to the array myEVENTFILE%WHICH_CCDNR
    ! the dimensions and the contents of the array  evttoolEVENT%ccdnr_infile
    ! which lists which CCD correaponds to what  CORRAT (ATT) extention
    ! the contents of the  evttoolEVENT%ccdnr_infile are inverse, 
    ! starting from higher and going to smaller. 
    ! CCDNR == CORRAT_EXTENTION_NUMBER BUT ATT INDEX IS INVERSE   
    ! i.e. CCDNR=6 -> CORRAR EXTENTION 6, BUT THIS IS STORED IN 
    ! ELEMENT 2 OF THE ARRAY ATT(2)
    allocate  ( myEVENTFILE%WHICH_CCDNR(evttoolEVENT%ninst  ) )
    myEVENTFILE%WHICH_CCDNR = evttoolEVENT%ccdnr_infile


    max_nrows = -9

    do ii = 1 , myEVENTFILE%NUM_ATT  
       
       nrows = evttoolEVENT%corrat(ii)%nelements
       
       if(nrows.gt.max_nrows) max_nrows = nrows

       myEVENTFILE%ATT(ii)%ACTIVE_TEL = evttoolEVENT%ccdnr_infile(ii)
       
       myEVENTFILE%ATT(ii)%SIZE = nrows
       
       allocate  ( myEVENTFILE%ATT(ii)%RA(myEVENTFILE%ATT(ii)%SIZE), &
            myEVENTFILE%ATT(ii)%DEC(myEVENTFILE%ATT(ii)%SIZE) )
       allocate  ( myEVENTFILE%ATT(ii)%ROLL(myEVENTFILE%ATT(ii)%SIZE), &
            myEVENTFILE%ATT(ii)%TIME(myEVENTFILE%ATT(ii)%SIZE) )
       
       
       myEVENTFILE%ATT(ii)%RA = evttoolEVENT%corrat(ii)%ra
       myEVENTFILE%ATT(ii)%DEC = evttoolEVENT%corrat(ii)%dec
       myEVENTFILE%ATT(ii)%ROLL = evttoolEVENT%corrat(ii)%roll
       myEVENTFILE%ATT(ii)%TIME = evttoolEVENT%corrat(ii)%time

       !write(*,*) "ATT:",evttoolEVENT%ccdnr_infile(ii), ii,evttoolEVENT%corrat(ii)%roll(1), evttoolEVENT%corrat(ii)%ccdnr, &
       !     evttoolEVENT%corrat(ii)%extname
    enddo


    call slib_init_ccd(myEVENTFILE)
    call slib_interpolate_roll(myEVENTFILE)

    ! sorting by spatial coordinates comes last because
    ! the estimation of roll angle assumes
    ! events are sorted by time.
    ! AGE 17.01.2019: this will be done outside
    ! the EVENT Initialisation task
    ! call slib_sort_heal(myEVENTFILE)

    
  end subroutine slib_evttool_2_myevent





  subroutine slib_init_ccd(myEVENTFILE)    
    ! this simply initialises the centre of each CCD
    ! in the future we may want to do more complex things
    ! i.e. read from header etc
    implicit none

    TYPE(EVENTFILE) :: myEVENTFILE

    integer irow, iactive


    !do iactive=1, myEVENTFILE%ATT%NUM_ACTIVE_TEL 
    do iactive=1, MAX_NUM_TELESCOPES

    
       myEVENTFILE%CCD%RAW_X0(iactive)=192.0
       myEVENTFILE%CCD%RAW_Y0(iactive)=192.0

    enddo

    
  end subroutine slib_init_ccd



  subroutine slib_interpolate_roll(myEVENTFILE)    
    
    implicit none

    TYPE(EVENTFILE) :: myEVENTFILE

    integer isrc, ii, iccdnr, left, right, jj
    
    double precision atime,x0, x1, y0, y1, y
 

    left = 0


    do isrc = 1, myEVENTFILE%EVENT%SIZE
       
       atime = myEVENTFILE%EVENT%TIME(isrc)
       iccdnr = myEVENTFILE%EVENT%CCDNR(isrc)
       
       ! find which element of array myEVENTFILE%WHICH_CCDNR
       ! the ccd number is
       do jj = 1, SIZE(myEVENTFILE%WHICH_CCDNR)
          if( myEVENTFILE%WHICH_CCDNR(jj) == iccdnr ) ii=jj
       enddo


       !write(*,*)  myEVENTFILE%WHICH_CCDNR
       !write(*,*) "1",isrc, atime, iccdnr, ii, myEVENTFILE%WHICH_CCDNR(ii)
       !where ( (myEVENTFILE%ATT(ii)%TIME - atime) .ge. 0 ) 
       !   left=minloc((myEVENTFILE%ATT(ii)%TIME - atime),1)
       !ENDWHERE
       !left =  minloc(abs(myEVENTFILE%ATT(ii)%TIME - atime),1)
       !if(left<1)left=1
       !if(left>=myEVENTFILE%ATT(ii)%SIZE)left=myEVENTFILE%ATT(ii)%SIZE-1
       !right =left+1
       
       call slib_find_bracket (myEVENTFILE, atime, ii, left, right )
       
       !write(*,*) "left,right", left, right
       !fix GL 29/08/2019
       left=max(1,left)
       y0=myEVENTFILE%ATT(ii)%ROLL(left) 
       y1=myEVENTFILE%ATT(ii)%ROLL(right)
       x0=myEVENTFILE%ATT(ii)%TIME(left) 
       x1=myEVENTFILE%ATT(ii)%TIME(right)

       !write(*,*) "3", x0, atime, x1, y0, y1


       y = y0 + (y1 - y0) * (atime - x0) / (x1 - x0)

       myEVENTFILE%EVENT%ROLL(isrc)=y !* myRAD2DEG

       !write(*,*) "ASSIGN:",isrc, iccdnr, ii, x0, atime, x1, "   ", y0, y, y1

    enddo

    
  end subroutine slib_interpolate_roll



  subroutine slib_find_bracket (myEVENTFILE, atime, itelesc, left, right )
    
    implicit none

    TYPE(EVENTFILE) :: myEVENTFILE

    integer irow, itelesc, left, right
    
    double precision atime
    
    do irow = left + 1, myEVENTFILE%ATT(itelesc)%SIZE - 1
       
       if ( atime < myEVENTFILE%ATT(itelesc)%TIME(irow) ) then
          left = irow - 1
          right = irow
          return
       end if

    end do
    
!    left = myEVENTFILE%ATT%SIZE(itelesc) - 1
!    right = myEVENTFILE%ATT%SIZE(itelesc)
    left = myEVENTFILE%ATT(itelesc)%SIZE - 1
    right = myEVENTFILE%ATT(itelesc)%SIZE
   
  end subroutine slib_find_bracket

  

  subroutine slib_sort_heal(myEVENTFILE)

    use slib_heal
    
    implicit none

    TYPE(EVENTFILE) :: myEVENTFILE

    integer, dimension (:), allocatable :: SINDECES ! sorted indeces
    integer ii, ipix_temp, var
    
    allocate (SINDECES(myEVENTFILE%EVENT%SIZE))
    SINDECES=0
    ! this is healpix - sorting pixels by ra and dec in healpix
    ! returns indeces that sort IPIX array
    call sortPIX(myEVENTFILE%EVENT%IPIX, SINDECES)

    ! now rearrange all arrays by those indices
    myEVENTFILE%EVENT%RA = myEVENTFILE%EVENT%RA(SINDECES)
    myEVENTFILE%EVENT%DEC = myEVENTFILE%EVENT%DEC(SINDECES)
    myEVENTFILE%EVENT%RAWX = myEVENTFILE%EVENT%RAWX(SINDECES)
    myEVENTFILE%EVENT%RAWY = myEVENTFILE%EVENT%RAWY(SINDECES)
    myEVENTFILE%EVENT%ENERGY = myEVENTFILE%EVENT%ENERGY(SINDECES)
    myEVENTFILE%EVENT%TIME = myEVENTFILE%EVENT%TIME(SINDECES)
    myEVENTFILE%EVENT%CCDNR = myEVENTFILE%EVENT%CCDNR(SINDECES)
    myEVENTFILE%EVENT%XD = myEVENTFILE%EVENT%XD(SINDECES)
    myEVENTFILE%EVENT%YD = myEVENTFILE%EVENT%YD(SINDECES)
    myEVENTFILE%EVENT%ZD = myEVENTFILE%EVENT%ZD(SINDECES)
    myEVENTFILE%EVENT%ROLL = myEVENTFILE%EVENT%ROLL(SINDECES)
    myEVENTFILE%EVENT%EV2SRC = myEVENTFILE%EVENT%EV2SRC(SINDECES)

    
    ipix_temp=-9    
    do ii=1, myEVENTFILE%EVENT%SIZE

       ! populating indeces for spatial ordering
       if(ipix_temp.ne.myEVENTFILE%EVENT%IPIX(ii))then	
          ipix_temp=myEVENTFILE%EVENT%IPIX(ii)
          myEVENTFILE%EVENT%INDEX1(ipix_temp)=ii
       endif
       myEVENTFILE%EVENT%INDEX2(ipix_temp)=ii

 !      write(*,*) ii, myEVENTFILE%EVENT%IPIX(ii),&
 !           myEVENTFILE%EVENT%RA(ii), myEVENTFILE%EVENT%DEC(ii), &
 !           myEVENTFILE%EVENT%RAWX(ii),myEVENTFILE%EVENT%RAWY(ii), &
 !           myEVENTFILE%EVENT%ENERGY(ii),  myEVENTFILE%EVENT%TIME(ii),&
 !           myEVENTFILE%EVENT%CCDNR(ii)
    enddo
    
    ! update flag that soreting by position
    ! has been done
    myEVENTFILE%EVENT%HEALSORT = .TRUE.

    ! deallocate SINDECES
    DEALLOCATE (SINDECES)

  end subroutine slib_sort_heal
    
end  module slib_data_evt
