! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module slib_heal

  use slib_constants
  use pix_tools
  
  implicit none
  private
  public:: SkyToPixTable, PixToSkyTable, sortPIX, NumPixels, GetPixRad, SkyToPix

  

contains

  subroutine NumPixels(nside, numpix)
    implicit none

    integer, intent(in) :: nside
    integer, intent(out) :: numpix
    
    numpix=nside2npix(nside)
    
  end subroutine NumPixels

  
  subroutine PixToSky(ra, dec, nside, ipnest)

    implicit none

    integer, intent(in) :: nside
    integer, intent(in) :: ipnest
    double precision, intent(out) :: ra, dec

    double precision theta, phi
        
    call pix2ang_nest(nside, ipnest, theta, phi)
    
    dec = ( -theta + myPI / 2.0 ) * myRAD2DEG
    ra = (2.0 * myPI - phi ) * myRAD2DEG

  end subroutine PixToSky

  
  subroutine SkyToPix(ra, dec, nside, ipnest)

    implicit none

    integer, intent(in) :: nside
    integer, intent(out) :: ipnest
    double precision, intent(in) :: ra, dec

    double precision theta, phi

    theta = (-dec+90.0) * myDEG2RAD
    phi = (360.0-ra) * myDEG2RAD
        
    call ang2pix_nest(nside, theta, phi, ipnest)


  end subroutine SkyToPix




  subroutine SkyToPixTable(RA, DEC, NSIDE, IPIX)

    implicit none

    integer, intent(in) :: NSIDE
    integer, intent(out) :: IPIX(:)
    double precision, intent(in) :: RA(:), DEC(:)

    integer I
    
    do I = 1, SIZE(RA)

       call SkyToPix(RA(I), DEC(I), NSIDE, IPIX(I))
       
    enddo
    
  end subroutine SkyToPixTable



  subroutine GetPixRad(ipix, search_radius_rad, nside, listpix, nlist)

    implicit none 

    double precision, intent(in) :: search_radius_rad
    integer, intent(in) :: ipix
    integer, intent(in) :: nside
    integer, intent(out), dimension(:) :: listpix 
    integer, intent(out) :: nlist

    integer nest

    double precision, dimension(3) :: vector 
    !double precision :: theta, phi
    integer i

    !call pix2ang_nest(nside, ipix, theta, phi)
!    write(*,*) i, ipix, phi/deg2rad, (theta-pi_trig/2.0)/deg2rad, search_radius_rad

    call pix2vec_nest(nside, ipix, vector)
    
    !write(*,*) nside_healpix_regions, ipix, search_radius_rad, SIZE(vector)   
    
    !do i=1,SIZE(vector)
    !   write(*,*) vector(i)
    !enddo

    call query_disc(nside, vector, search_radius_rad, &
         listpix, nlist, nest=1, inclusive=1) 

!    do i=1,nlist
!       call pix2ang_nest(nside_healpix_regions, listpix(i), theta, phi)
!       write(*,*) i, listpix(i), phi/deg2rad, (theta-pi_trig/2.0)/deg2rad, search_radius_rad      
!    enddo

  end subroutine GetPixRad
  

  

  subroutine PixToSkyTable(RA, DEC, NSIDE, IPIX)

    implicit none

    integer, intent(in) :: NSIDE
    integer, intent(in) :: IPIX(:)
    double precision, intent(out) :: RA(:), DEC(:)

    integer I
    
    do I = 1, SIZE(IPIX)

       call PixToSky(RA(I), DEC(I), NSIDE, IPIX(I))
       
    enddo
    
  end subroutine PixToSkyTable


  

  subroutine sortPIX(ipix, ind)
    
    implicit none
    
    integer, intent(inout) :: ipix(:)
    integer, intent(out) :: ind(:)
    
    integer :: n, i, ir, j, l, iind  
    integer :: iipix

    !ind = 0
    n=SIZE(ind)
    ! initialize index array
    IF (ind(1) .eq.0) then  
       DO i = 1, n  
          ind(i) = i
          !write(*,*) "HERE",i,ind(i)
       ENDDO
    ENDIF
    !write(*,*)
    
    ! nothing to order
    IF (n.lt.2) return  
    ! initialize indices for hiring and retirement-promotion phase
    l = n / 2 + 1  
    ir = n  
    !write(*,*) n, l, ir
  sorting: do 
     ! still in hiring phase
     IF ( l .gt. 1 ) then  
        l    = l - 1  
        iipix  = ipix (l)  
        iind = ind (l)  
       ! in retirement-promotion phase.
     ELSE  
        ! clear a space at the end of the array
        iipix  = ipix (ir)  
        !
        iind = ind (ir)  
        ! retire the top of the heap into it
        ipix (ir) = ipix (1)  
        !
        ind (ir) = ind (1)  
        ! decrease the size of the corporation
        ir = ir - 1  
        ! done with the last promotion
        IF ( ir .eq. 1 ) then  
           ! the least competent worker at all !
          ipix (1)  = iipix  
          !
          ind (1) = iind  
          exit sorting  
       ENDIF
    ENDIF
    ! wheter in hiring or promotion phase, we
    i = l  
    ! set up to place rra in its proper level
    j = l + l  
    !
    DO while(j.le.ir)
       if(j.lt.ir)then
          if ( ipix(j) < ipix(j+1) )then
             j = j +1
          endif
       endif

       if( iipix<ipix(j) ) then

          ipix(i) = ipix(j)
          ind(i)= ind(j)
          i=j
          j=j+j
       else
          j=ir+1
       endif
    enddo
    
    ipix(i)=iipix
    ind(i)=iind
    !write(*,*) i, ipix(i), ind(i)
 end do sorting
    
  end subroutine sortPIX
  
end module slib_heal
