! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module slib_image

  use slib_constants
  
  implicit none
  
  private
  
  public :: slib_run_write2arrays_fast, slib_run_write2image, slib_recon_inimage, slib_fluxR_image


contains 



  subroutine slib_fluxR_image(myCPSF, M, R, FLUXR)
    implicit none

    TYPE(CPSF), intent(in) :: myCPSF
    INTEGER, intent(in) :: M
    DOUBLE PRECISION, DIMENSION(M),  intent(in) :: R
    DOUBLE PRECISION, DIMENSION(M),  intent(out) :: FLUXR

    DOUBLE PRECISION, DIMENSION(M) :: R2

    DOUBLE PRECISION xo, yo, dr2, x, y
    integer ix, iy, ir
    
    ! center of PSF image    
    FLUXR=0.0
    xo=myCPSF%MAX_PSF_IMG_SIZE/2.0+1.0
    yo=myCPSF%MAX_PSF_IMG_SIZE/2.0+1.0
    R2= R * R

    !write(*,*) xo, yo
    !write(*,*) R2
    do ix=1,myCPSF%MAX_PSF_IMG_SIZE
       x=real(ix)          
       do iy=1,myCPSF%MAX_PSF_IMG_SIZE
          y=real(iy)

          dr2=(x-xo) * (x-xo) + (y-yo) * (y-yo)

          do ir=1,M
             if(dr2.le.R2(ir))then
                FLUXR(ir)=FLUXR(ir) + myCPSF%IMG(ix,iy)
             endif
          enddo
          
       enddo
    enddo

    
  end subroutine slib_fluxR_image

  subroutine slib_run_write2arrays_fast(myStackPSF, myCPSF, ISRC)

    implicit none

    TYPE(STACKPSF) :: myStackPSF
    TYPE(CPSF) :: myCPSF
    INTEGER, intent(in) ::  ISRC

    
    INTEGER N_COEFF0, N_COEFF1, N_COEFF2    
    INTEGER i 
    DOUBLE PRECISION A1

    N_COEFF0 = SIZE(myCPSF%NX0)
    N_COEFF1 = SIZE(myCPSF%NX1)
    N_COEFF2 = SIZE(myCPSF%NX2)

    myCPSF%IMG=0.0

    !write(*,*) N_COEFF0, ISRC, "SHAPE:",SHAPE(myStackPSF%C0)

    do i = 1, N_COEFF0

       !write(*,*) i, N_COEFF0, ISRC

        A1= myStackPSF%C0(i, ISRC)
        
        myCPSF%IMG(:,:) = myCPSF%IMG(:,:) + A1 * myCPSF%IMG_SHAPE0(:,:, i) 
      ! write(*,*) myCPSF%IMG(:,:)

     enddo

     
     do i = 1, N_COEFF1

       !write(*,*) i, N_COEFF1, ISRC, A1

        A1= myStackPSF%C1(i, ISRC)
        
        myCPSF%IMG(:,:) = myCPSF%IMG(:,:) + A1 * myCPSF%IMG_SHAPE1(:,:, i) 
      ! write(*,*) myCPSF%IMG(:,:)

     enddo
     
     do i = 1, N_COEFF2

        A1= myStackPSF%C2(i, ISRC)
        
        myCPSF%IMG(:,:) = myCPSF%IMG(:,:) + A1 * myCPSF%IMG_SHAPE2(:,:, i) 

     enddo
        
   end subroutine slib_run_write2arrays_fast




  subroutine slib_run_write2arrays(myStackPSF, myCPSF, ISRC)

    use slib_basic_math

    implicit none


    TYPE(STACKPSF) :: myStackPSF
    TYPE(CPSF) :: myCPSF

    INTEGER ISRC

  
    DOUBLE PRECISION, DIMENSION(0:SIZE(myCPSF%NX0)) :: cx0,cy0  
    DOUBLE PRECISION, DIMENSION(0:SIZE(myCPSF%NX1)) :: cx1,cy1      
    DOUBLE PRECISION, DIMENSION(0:SIZE(myCPSF%NX2)) :: cx2,cy2  
    
    INTEGER N_COEFF0, N_COEFF1, N_COEFF2
    
    DOUBLE PRECISION BETA0, BETA1, BETA2

    
    DOUBLE PRECISION x, y, xo, yo, xprime, yprime, pixsize
    
    INTEGER i, ix, iy, ix_hermite, iy_hermite
    
    DOUBLE PRECISION normx, normy, A1, SUBSAMPLE


    SUBSAMPLE = RAW_PIX_MICRON / myCPSF%IMG_PIXEL_MM

    i = int(SUBSAMPLE * 100)

    if( myCPSF%MAX_PSF_IMG_SIZE < i  )then
       write(*,*) "WARNING: PSF RESOLUTION SUBSAMPLED BY FACTOR", SUBSAMPLE 
       write(*,*) "WARNING: OUTPUT ARRAY SIZE:", myCPSF%MAX_PSF_IMG_SIZE," MAY BE TOO SMALL"
       write(*,*) "YOU MAY WANT TO CHANGE PARAMETER MAX_PSF_IMG_SIZE OR IMG_PIXEL_MM"
    endif


    ! center of PSF image    
    xo=myCPSF%MAX_PSF_IMG_SIZE/2.0+1.0
    yo=myCPSF%MAX_PSF_IMG_SIZE/2.0+1.0
    myCPSF%IMG=0.0
    pixsize=1
    N_COEFF0 = SIZE(myCPSF%NX0)
    N_COEFF1 = SIZE(myCPSF%NX1)
    N_COEFF2 = SIZE(myCPSF%NX2)
    BETA0  = myCPSF%BETA0 * SUBSAMPLE
    BETA1  = myCPSF%BETA1 * SUBSAMPLE
    BETA2  = myCPSF%BETA2 * SUBSAMPLE


    !write(*,*) LOOKUP_N1
    !write(*,*) LOOKUP_N2
    !write(*,*) C_COEFF


    i=1
    do ix=1,myCPSF%MAX_PSF_IMG_SIZE
       x=real(ix)*pixsize-pixsize/2.
          
       do iy=1,myCPSF%MAX_PSF_IMG_SIZE
          y=real(iy)*pixsize-pixsize/2.



          ! this is because NORTH is always aligned with
          ! ERO-X and EAST with ERO-Y. I need to create images
          ! with NORTH==+X, EAST=-Y
          xprime=(y-yo)
          yprime=(x-xo)
          !xprime=(x-xo)
          !yprime=(y-yo)
          xprime=xprime/BETA0
          yprime=yprime/BETA0
             
          !write(*,*) x,y,ix,iy,xprime,yprime,beta, NIMG
             
          call hermite_poly ( N_COEFF0, dble(xprime), cx0 )
          call hermite_poly ( N_COEFF0, dble(yprime), cy0 )
             
          do i = 1, N_COEFF0
             ix_hermite=myCPSF%NX0(I)
             iy_hermite=myCPSF%NY0(I)
                
             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
                
             normx=1. / sqrt( 2**ix_hermite * sqrt(myPI) * GAMMA(real(ix_hermite+1)) ) !factrl(ix_hermite) )
             normx=normx*1./sqrt(BETA0)*exp(-xprime**2/2)  
             normy=1. / sqrt( 2**iy_hermite * sqrt(myPI) * GAMMA(real(iy_hermite+1)) ) !factrl(iy_hermite) )
             normy=normy*1./sqrt(BETA0)*exp(-yprime**2/2)  
             
             !if(ix==10 && iy==4)then
             !   write(*,*) normx, 2**ix_hermite, sqrt(myPI), GAMMA(real(ix_hermite+1)), exp(-xprime**2/2) , xprime, sqrt(BETA)
             !enddo
             A1= myStackPSF%C0(i, ISRC)
             myCPSF%IMG(ix,iy) = myCPSF%IMG(ix,iy) + &
                  A1 * normx *normy * real(cx0(ix_hermite)) * real(cy0(iy_hermite))
                
                
          enddo


          
          ! this is because NORTH is always aligned with
          ! ERO-X and EAST with ERO-Y. I need to create images
          ! with NORTH==+X, EAST=-Y
          xprime=(y-yo)
          yprime=(x-xo)
          !xprime=(x-xo)
          !yprime=(y-yo)
          xprime=xprime/BETA1
          yprime=yprime/BETA1
             
          !write(*,*) x,y,ix,iy,xprime,yprime,beta, NIMG
             
          call hermite_poly ( N_COEFF1, dble(xprime), cx1 )
          call hermite_poly ( N_COEFF1, dble(yprime), cy1 )
             
          do i = 1, N_COEFF1
             ix_hermite=myCPSF%NX1(I)
             iy_hermite=myCPSF%NY1(I)
                
             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
                
             normx=1. / sqrt( 2**ix_hermite * sqrt(myPI) * GAMMA(real(ix_hermite+1)) ) !factrl(ix_hermite) )
             normx=normx*1./sqrt(BETA1)*exp(-xprime**2/2)  
             normy=1. / sqrt( 2**iy_hermite * sqrt(myPI) * GAMMA(real(iy_hermite+1)) ) !factrl(iy_hermite) )
             normy=normy*1./sqrt(BETA1)*exp(-yprime**2/2)  
             
             !if(ix==10 && iy==4)then
             !   write(*,*) normx, 2**ix_hermite, sqrt(myPI), GAMMA(real(ix_hermite+1)), exp(-xprime**2/2) , xprime, sqrt(BETA)
             !enddo
             A1= myStackPSF%C1(i, ISRC)
             myCPSF%IMG(ix,iy) = myCPSF%IMG(ix,iy) + &
                  A1 * normx *normy * real(cx1(ix_hermite)) * real(cy1(iy_hermite))
                
                
          enddo
             
          ! this is because NORTH is always aligned with
          ! ERO-X and EAST with ERO-Y. I need to create images
          ! with NORTH==+X, EAST=-Y
          xprime=(y-yo)
          yprime=(x-xo)
          !xprime=(x-xo)
          !yprime=(y-yo)
          xprime=xprime/BETA2
          yprime=yprime/BETA2
          
          !write(*,*) x,y,ix,iy,xprime,yprime,beta, NIMG
          
          call hermite_poly ( N_COEFF2, dble(xprime), cx2 )
          call hermite_poly ( N_COEFF2, dble(yprime), cy2 )
          
          do i = 1, N_COEFF2
             ix_hermite=myStackPSF%NX2(I)
             iy_hermite=myStackPSF%NY2(I)
                
             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
             
             normx=1. / sqrt( 2**ix_hermite * sqrt(myPI) * GAMMA(real(ix_hermite+1)) ) !factrl(ix_hermite) )
             normx=normx*1./sqrt(BETA2)*exp(-xprime**2/2)  
             normy=1. / sqrt( 2**iy_hermite * sqrt(myPI) * GAMMA(real(iy_hermite+1)) ) !factrl(iy_hermite) )
             normy=normy*1./sqrt(BETA2)*exp(-yprime**2/2)  
             
             !if(ix==10 && iy==4)then
             !   write(*,*) normx, 2**ix_hermite, sqrt(myPI), GAMMA(real(ix_hermite+1)), exp(-xprime**2/2) , xprime, sqrt(BETA)
             !enddo
             A1= myStackPSF%C2(i, ISRC)
             myCPSF%IMG(ix,iy) = myCPSF%IMG(ix,iy) + &
                  A1 * normx *normy * real(cx2(ix_hermite)) * real(cy2(iy_hermite))
             
             !write(*,*) ix_hermite, iy_hermite, normx, normy,  A1 ,  real(cx(ix_hermite)),  real(cy(iy_hermite)),  array(ix,iy)
                
          enddo
             
          !write(*,*) ix, iy, array(ix,iy)
             
       enddo

    enddo
       
    
  end subroutine slib_run_write2arrays





 subroutine slib_run_write2image(ARRAY, filename)

    DOUBLE PRECISION, DIMENSION(:,:) :: ARRAY
    !TYPE(STACKPSF) :: myStackPSF
    character ( len = *) filename


    !Fitsio needs these
    integer :: status,unit,blocksize,bitpix,naxis
    integer, DIMENSION(2) :: naxes(2)
    integer ::  nhdu,hdutype,group,fpixel,nelements,nrows
    logical simple,extend    
    !DOUBLE PRECISION, DIMENSION(myStackPSF%MAX_PSF_IMG_SIZE,myStackPSF%MAX_PSF_IMG_SIZE) :: ARRAY

       
    !ARRAY = myStackPSF%IMG

    ! WRITE_IMAGE
    naxes = SHAPE(ARRAY)

    status=0
       
    !  Name of the FITS file to be created:
    !write (filename, "(I4.4, A2, I4.4)") &
    !     ISRC, "_E", IE
    !   filename=name(1:iname)//"_SRC"//trim(filename)//'.img'


    !  Delete the file if it already exists, so we can then recreate it.
    call delete_file ( filename, status )
       
    !  Get an unused Logical Unit Number to use to open the FITS file.
    call ftgiou ( unit, status )
    !
    !  Create the new empty FITS file.
    !
    blocksize=1
    call ftinit(unit,filename,blocksize,status)
    !
    !  Initialize parameters about the FITS image (300 x 200 16-bit integers).
    !
    simple=.true. 
    bitpix=-32
    naxis=2
    !naxes(1)=myStackPSF%MAX_PSF_IMG_SIZE
    !naxes(2)=myStackPSF%MAX_PSF_IMG_SIZE
    extend=.true.
       
    !  Write the required header keywords.
    call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)
    
    !  Write the array to the FITS file.
       
    group=1
    fpixel=1
    nelements=naxes(1)*naxes(2)
    !call ftpprd(unit,group,fpixel,nelements,array_gauss,status)
    call ftpprd(unit,group,fpixel,nelements,ARRAY,status)
    !
    !  Write another optional keyword to the header.
    !
    !  call ftpkyj(unit,'EXPOSURE',1500,'Total Exposure Time',status)
    !
    !  Close the file and free the unit number.
    !
    call ftclos(unit, status)
    call ftfiou(unit, status)
    !
    !  Check for any error, and if so print out error messages
    !
    !if (status > 0) then
    !   call print_error(status)
    !end if

    
  end subroutine slib_run_write2image


 subroutine slib_recon_inimage(imafile, myPSF, myStackPSF, mySRC)
   use slib_basic_math
   use fitsimage

   TYPE(STACKPSF) :: myStackPSF
   TYPE(CPSF) :: myPSF
   TYPE(SRC) :: mySRC
   INTEGER ISRC

   
   DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX0)) :: cx0,cy0      
   DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX1)) :: cx1,cy1      
   DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX2)) :: cx2,cy2  
   INTEGER N_COEFF0, N_COEFF1, N_COEFF2
   DOUBLE PRECISION A1,BETA0,BETA1, BETA2, f
   DOUBLE PRECISION x, y, xo, yo
   INTEGER i, ix, iy, ideltaX, ideltaY, ix_hermite, iy_hermite
   
   character(len=*) :: imafile
   type(shapehdr_type) :: shapehdr
   type(stdhdr_type) :: stdhdr
   type(wcs_refhdr_type) :: refhdr
   type(wcs_imahdr_type) :: wcshdr
   

   !Fitsio needs these
   integer :: unit, ierr = 0, status = 0, ftread = 0
   integer :: blocksize,bitpix,naxis
   integer, DIMENSION(2) :: naxes(2)
   integer ::  nhdu,hdutype,group,fpixel,nelements,nrows
   logical simple,extend    
   character(len=30) :: comment
   
   DOUBLE PRECISION, DIMENSION(:,:), allocatable :: ARRAY

   
   ! get units and camera numbers from event file
   ! this is taken from the exposure map routines in expmap_io_mod
   !call opn_events(1, imafile, ncamarr, camarr, camlun, instarr, ierr) 
   !if (ierr > 2) call wrstop(rtname, &
   !      "Coulnd't obtain camera numbers")

   ! get keywords from template image ===============================
    call ftgiou (unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftiopn(unit,imafile, ftread, status) 
    call fitserr("ftiopn", status, ierr)
    !if (status > 0) call wrstop(rtname, &
    !     "Couldn't open template image")
    ! the calls (eroio - fitsimage) below essentially read the
    ! image extension of the event file 
    status = 0
    ! get image shape headers
    call fread_shapehdr(unit, shapehdr, ierr) 
    ! get eSASS standard headers
    call fread_stdhdr(unit, stdhdr, ierr)
    ! get wcs keywords
    call fread_refhdr(unit, refhdr, ierr)
    call fread_wcshdr(unit, wcshdr, ierr)
    call ftclos(unit, status)
    call ftfiou(unit, status)

    ! write array
    write(*,*) shapehdr%naxis1,shapehdr%naxis2
    allocate  ( ARRAY(shapehdr%naxis1,shapehdr%naxis2) )
    N_COEFF0 = SIZE(myPSF%NX0)
    N_COEFF1 = SIZE(myPSF%NX1)
    N_COEFF2 = SIZE(myPSF%NX2)
    BETA0  = myPSF%BETA0
    BETA1  = myPSF%BETA1
    BETA2  = myPSF%BETA2
    f = 1./ (RAW_PIX_MICRON / myPSF%IMG_PIXEL_MM)
    ARRAY=0.0

    !write(*,*) BETA0,BETA1,BETA2

    do ISRC=1,size(mySRC%XIMA)
       !write(*,*) "HERE",ISRC, size(mySRC%XIMA)
       xo=mySRC%XIMA(ISRC)
       yo=mySRC%YIMA(ISRC)

       write(*,*) "Reconstructing",ISRC, xo, yo

       ! center of pixels is integer
       do ideltaX=-50,50
          x = xo + real(ideltaX)
          ix = int(x)
       
          do ideltaY=-50,50
             y = yo + real(ideltaY)
             iy=int(y)
             
             !write(*,*) "IMG:",x, y, xo, yo, ix, iy, f
             !write(*,*) "RECON:",f*(y-0.5), f*(y+0.5), f*(x-0.5), f*(x+0.5), f*yo, f*xo


             ! this is because NORTH is always aligned with
             ! ERO-X and EAST with ERO-Y. I need to create images
             ! with NORTH==+X, EAST=-Y
             call  hermite_integrate2D(N_COEFF0, BETA0, f*(y-0.5), f*(y+0.5), f*(x-0.5), f*(x+0.5), f*yo, f*xo, cx0, cy0)
             do i = 1, N_COEFF0
                ix_hermite=myPSF%NX0(I)
                iy_hermite=myPSF%NY0(I)
                A1= myStackPSF%C0(i, ISRC)
                !write(*,*) x, y, ix_hermite, iy_hermite, cx1(ix_hermite), cy1(iy_hermite), xo, yo
                ARRAY(ix,iy) =   ARRAY(ix,iy) + A1 * cx0(ix_hermite) * cy0(iy_hermite)
             enddo



             
             ! this is because NORTH is always aligned with
             ! ERO-X and EAST with ERO-Y. I need to create images
             ! with NORTH==+X, EAST=-Y
             call  hermite_integrate2D(N_COEFF1, BETA1, f*(y-0.5), f*(y+0.5), f*(x-0.5), f*(x+0.5), f*yo, f*xo, cx1, cy1)


             do i = 1, N_COEFF1
                ix_hermite=myPSF%NX1(I)
                iy_hermite=myPSF%NY1(I)
                A1= myStackPSF%C1(i, ISRC)
                !write(*,*) x, y, ix_hermite, iy_hermite, cx1(ix_hermite), cy1(iy_hermite), xo, yo
                ARRAY(ix,iy) =   ARRAY(ix,iy) + A1 * cx1(ix_hermite) * cy1(iy_hermite)
             enddo


             ! this is because NORTH is always aligned with
             ! ERO-X and EAST with ERO-Y. I need to create images
             ! with NORTH==+X, EAST=-Y
             call  hermite_integrate2D(N_COEFF2, BETA2, f*(y-0.5), f*(y+0.5), f*(x-0.5), f*(x+0.5), f*yo, f*xo, cx2, cy2)
          
             do i = 1, N_COEFF2
                ix_hermite=myPSF%NX2(I)
                iy_hermite=myPSF%NY2(I)
                A1= myStackPSF%C2(i, ISRC)
                ARRAY(ix,iy) =   ARRAY(ix,iy) + A1 * cx2(ix_hermite) * cy2(iy_hermite)
             enddo                      
          enddo
       enddo
    enddo

    ! write the thing
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftinit(unit, "!test", 0, status)
    call fitserr("ftinit", status, ierr)

    simple=.true. 
    bitpix=-32
    naxis=2
    naxes(1)=shapehdr%naxis1
    naxes(2)=shapehdr%naxis2
    extend=.true.
    !  Write the required header keywords.
    call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)

    call fwrite_stdhdr(unit,stdhdr,ierr)
    call fwrite_refhdr(unit,refhdr,ierr)
    call fwrite_wcshdr(unit,wcshdr,ierr)


    group=1
    fpixel=1
    nelements=naxes(1)*naxes(2)
    !call ftpprd(unit,group,fpixel,nelements,array_gauss,status)
    call ftpprd(unit,group,fpixel,nelements,ARRAY,status)

    call ftclos(unit, status)
    call ftfiou(unit, status)
        
    deallocate(ARRAY)
 end subroutine slib_recon_inimage

subroutine delete_file(filename,status)
!
!*******************************************************************************
!
!! DELETE_FILE deletes a FITS file.
!
  integer status,unit,blocksize
  character ( len = * ) filename
!
!  Simply return if status is greater than zero.
!

  !write(*,*) filename, status


  if (status > 0) then
    return
  end if
!
!  Get an unused Logical Unit Number to use to open the FITS file
!
  call ftgiou ( unit, status )
!
!  Try to open the file, to see if it exists
!
  call ftopen ( unit, filename, 1, blocksize, status )

  !write(*,*) unit, filename, blocksize, status

  if ( status == 0 ) then
!
!  File was opened;  so now delete it 
!
    call ftdelt(unit,status)

  else if (status == 103)then
!
!     File doesn't exist, so just reset status to zero and clear errors
!
    status=0
    call ftcmsg

  else
!
!  There was some other error opening the file; delete the file anyway
!
      status=0
      call ftcmsg
      call ftdelt(unit,status)
  end if
!
!  Free the unit number for later reuse.
!
  call ftfiou(unit, status)

  return
end subroutine delete_file
!****************************






end module slib_image
